import numpy as np
from igimf import util

class Plots:
    def __init__(self, dir_path='figs'):
        import os
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
            
    # Plotting functions
    def Migal_plot(self, M_igal_v, SFR, downsizing_time):
        from matplotlib import pyplot as plt
        import matplotlib.ticker as ticker
        Msun = r'$M_{\odot}$'
        fig, ax1 = plt.subplots(1,1, figsize=(7,5))
        ax0 = ax1.twinx()
        ax0.loglog(M_igal_v, SFR, linewidth=3, color='tab:red')
        ax0.set_ylabel(f'SFR [{Msun}/yr]', fontsize=15, color='tab:red')
        ax0.set_xlabel(r'$M_{igal}$ '+f'[{Msun}]', fontsize=15)
        ax1.semilogx(M_igal_v, downsizing_time, linewidth=3, color='tab:blue')
        ax1.set_ylabel(r'$\Delta\tau$ [Gyr]', fontsize=15, color='tab:blue')
        ax1.set_xlabel(r'$M_{igal}$ '+f'[{Msun}]', fontsize=15)
        #ax.set_ylim(1e-8,1)
        ax0.tick_params(width=2, axis='both', labelsize=15)
        ax1.tick_params(width=2, axis='both', labelsize=15)
        fig.tight_layout()
        #plt.savefig(f'figs/Z_plot_{name}.pdf', bbox_inches='tight')
        plt.show(block=False)
        return None
        
    def ECMF_plot(self, Mecl_v, ECMF_v, SFR):
        from matplotlib import pyplot as plt
        import matplotlib.ticker as ticker
        Msun = r'$M_{\odot}$'
        fig, ax = plt.subplots(1,1, figsize=(6,4))
        ax.loglog(Mecl_v, ECMF_v, linewidth=3, color='navy')
        ax.scatter(Mecl_v, ECMF_v, linewidth=3, color='navy',s=1)
        ax.set_ylabel(r'$\xi_{ECMF}$'+f' [#/{Msun}]', fontsize=15)
        ax.set_xlabel(r'$M_{\rm ecl}$ [%s]'%(Msun), fontsize=15)
        plt.title(r'$\,$ SFR = %.2e [%s/yr]' %(SFR, Msun), fontsize=15)
        #ax.set_ylim(1e-8,1)
        plt.yticks(fontsize=15)
        plt.xticks(fontsize=15)
        ax.tick_params(width=2)
        fig.tight_layout()
        plt.savefig(f'figs/ECMF_plot_SFR{SFR:.2e}.pdf', bbox_inches='tight')
        #plt.show(block=False)
            
    def beta_ECMF_bySFR_plot(self, SFR_v, beta_ECMF_v):
        from matplotlib import pyplot as plt
        import matplotlib.ticker as ticker
        Msun = r'$M_{\odot}$'
        fig, ax = plt.subplots(1,1, figsize=(6,4))
        ax.semilogx(SFR_v, beta_ECMF_v, linewidth=3, color='navy')
        ax.scatter(SFR_v, beta_ECMF_v, linewidth=3, color='navy',s=1)
        ax.set_ylabel(r'$\beta_{ECMF}$', fontsize=15)
        ax.set_xlabel(r'SFR [%s/yr]'%(Msun), fontsize=15)
        #ax.set_title(r'[Z] = %.2f' %(Z), fontsize=15)
        #ax.set_ylim(1e-8,1)
        plt.yticks(fontsize=15)
        plt.xticks(fontsize=15)
        ax.tick_params(width=2)
        fig.tight_layout()
        plt.savefig(f'figs/beta_ECMF_bySFR.pdf', bbox_inches='tight')
        #plt.show(block=False)
               
    def MeclMax_bySFR_plot(self, SFR_v, MeclMax_list):
        from matplotlib import pyplot as plt
        import matplotlib.ticker as ticker
        Msun = r'$M_{\odot}$'
        fig, ax = plt.subplots(1,1, figsize=(6,4))
        ax.loglog(SFR_v, MeclMax_list, linewidth=3, color='navy')
        ax.scatter(SFR_v, MeclMax_list, linewidth=3, color='navy',s=1)
        ax.set_ylabel(r'$M_{\rm ecl,max}$ [%s]'%(Msun), fontsize=15)
        ax.set_xlabel(r'SFR [%s/yr]'%(Msun), fontsize=15)
        #ax.set_title(r'[Z] = %.2f' %(Z), fontsize=15)
        #ax.set_ylim(5e-2,1e8)
        #ax.set_xlim(1e-4,1e3)
        plt.yticks(fontsize=15)
        plt.xticks(fontsize=15)
        ax.tick_params(width=2)
        fig.tight_layout()
        plt.savefig(f'figs/MeclMax_bySFR_plot.pdf', bbox_inches='tight')
        #plt.show(block=False)
        
             
    def Meclmax_vs_SFR_observations(self, SFR_v, MeclMax_list, k_ML=0.01):
        import pandas as pd
        from matplotlib import pyplot as plt
        import matplotlib.ticker as ticker
        r13 = pd.read_csv('data/randriamanakoto13.dat', sep=',', comment='#', index_col=False)
        l02 = pd.read_csv('data/larsen02.dat', sep=';', comment='#', index_col=False)
        b08 = pd.read_csv('data/bastian08.dat', sep=';', comment='#', index_col=False)
        def mag_to_mass(mag_V):
            #mag_V = mag_K + 2
            Mecl_max = np.power(10, np.divide(4.79 - mag_V, 2.5)) * k_ML
            return Mecl_max
        mag_to_mass = np.vectorize(mag_to_mass)
        Msun = r'$M_{\odot}$'
        fig, ax = plt.subplots(1,1, figsize=(6,4))
        ax.loglog(SFR_v, MeclMax_list, linewidth=2, color='navy')
        print(f'{SFR_v=}')
        print(f'{MeclMax_list=}')
        ax.scatter((r13['SFR(Msun/yr)']), (mag_to_mass(r13['M_K_brightest(mag)']+2)), label='Randriamanakoto+13', marker='o')
        ax.scatter((b08['SFR(Msun/yr)']), (mag_to_mass(b08['M_V_brightest(mag)'])), label='Bastian08', marker='s')
        ax.scatter((l02['SFRdensity(Msun/yr/kpc^2)']*l02['A(kpc^2)']), (mag_to_mass(l02['M_V_brightest(mag)'])), label='Larsen02', marker='^')
        ax.scatter(SFR_v, MeclMax_list, linewidth=2, color='navy',s=1)
        ax.set_ylabel(r'$M_{\rm ecl,max}$ [%s]'%(Msun), fontsize=15)
        ax.set_xlabel(r'SFR [%s/yr]'%(Msun), fontsize=15)
        #ax.set_ylim(10**(-1.5),1e8)
        ax.set_xlim(1e-6,.5e3)
        #ax.set_xlim(1e-6, 1e6)
        plt.yticks(fontsize=15)
        plt.xticks(fontsize=15)
        ax.tick_params(width=2)
        fig.tight_layout()
        plt.savefig(f'figs/Meclmax_vs_SFR_observations.pdf', bbox_inches='tight')
        
        
    def Mecl_power_beta_plot(self, Mecl_v, beta_ECMF_list):
        from matplotlib import pyplot as plt
        import matplotlib.ticker as ticker
        import colorcet as cc
        from mpl_toolkits.axes_grid1 import make_axes_locatable
        Msun = r'$M_{\odot}$'
        cm = cc.cm.isolum
        num_colors = len(beta_ECMF_list)
        Z = [[0,0],[0,0]]
        beta_ECMF_list = np.flip(beta_ECMF_list)
        levels = np.linspace(np.min(beta_ECMF_list), np.max(beta_ECMF_list),
                             100, endpoint=True)
        CS3 = plt.contourf(Z, levels, cmap=cm)
        plt.clf()
        currentColors = [cm(1.*i/num_colors) for i in range(num_colors)]
        currentColor = iter(currentColors)
        fig, ax = plt.subplots(1,1, figsize=(7,5))
        for b in beta_ECMF_list:
            y = Mecl_v**(-b)
            color = next(currentColor)
            ax.loglog(Mecl_v, y, linewidth=3, c=color)
            #ax.scatter(Mecl_v, y, linewidth=3, c=color,s=1)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad="2%")
        ax.set_ylabel(r'$\xi_{ECMF}$'+f' [#/{Msun}]'+' not normalized',
                      fontsize=15)
        ax.set_xlabel(r'$M_{\rm ecl}$'+f' [{Msun}]', fontsize=15)
        ax.set_title(r'$k_{\rm ecl}=1$', fontsize=15)
        ax.set_ylim(1e-15,1e1)
        plt.yticks(fontsize=15)
        plt.xticks(fontsize=15)
        ax.tick_params(width=2)
        cbar = fig.colorbar(CS3, cmap=cm, cax=cax, format="%.2f", 
                            ticks=ticker.MultipleLocator(.2)).set_label(
                                label=r'$\beta_{\rm ECMF}$',size=15)
        fig.tight_layout()
        plt.savefig(f'figs/Mecl_power_beta.pdf', bbox_inches='tight')
        #plt.show(block=False)
        
    def ECMF_plots(self, M_ecl_v, ECMF_v_list, SFR_v):
        from matplotlib import pyplot as plt
        import matplotlib.ticker as ticker
        from mpl_toolkits.axes_grid1 import make_axes_locatable
        Msun = r'$M_{\odot}$'
        cm = plt.cm.get_cmap(name='inferno')
        num_colors = len(ECMF_v_list)
        Z = [[0,0],[0,0]]
        #levels = np.linspace(np.log10(SFR_v[0]), np.log10(SFR_v[-1]), num_colors, endpoint=True)
        levels = np.linspace(np.log10(SFR_v[0]), np.log10(SFR_v[-1]), num_colors, #109 
                             endpoint=True)
        CS3 = plt.contourf(Z, levels, cmap=cm)
        plt.clf()
        fig, ax = plt.subplots(1,1, figsize=(7,5))
        SFR_colormap = (SFR_v)#np.log10(np.logspace(np.log10(SFR[0]), np.log10(SFR[-1]), 10, endpoint=True))
        currentColors = [cm(1.*i/num_colors) for i in range(num_colors)]
        currentColor = iter(currentColors)
        for i,ECMF in enumerate(ECMF_v_list):
            ax.loglog(M_ecl_v,ECMF, linewidth=3, c=next(currentColor))
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad="2%")
        ax.set_ylabel(r'$\xi_{ECMF}$'+ f' [#/{Msun}]', fontsize=15)
        ax.set_xlabel(r'$M_{\rm ecl}$ [%s]' %(Msun), fontsize=15)
        #ax.set_ylim(1e-1,1e5)
        ax.tick_params(width=2)
        cbar = fig.colorbar(CS3, cmap=cm, cax=cax, format="%.2f", 
                            ticks=ticker.MultipleLocator(1)).set_label(label=r'$\log_{10}({\rm SFR})$'+f' [({Msun}/yr)]',size=15)
        fig.tight_layout()
        plt.savefig(f'figs/ECMF_plots.pdf', bbox_inches='tight')
        #plt.show(block=False)
        return None
  
    def gwIMF_plots(self, star_v, gwIMF_bySFR_eval, SFR_v, metal_mass_fraction):
        from matplotlib import pyplot as plt
        import matplotlib.ticker as ticker
        from mpl_toolkits.axes_grid1 import make_axes_locatable
        Msun = r'$M_{\odot}$'
        cm = plt.cm.get_cmap(name='magma')
        num_colors = len(gwIMF_bySFR_eval)
        Z = [[0,0],[0,0]]
        #levels = np.linspace(np.log10(SFR_v[0]), np.log10(SFR_v[-1]), num_colors, endpoint=True)
        levels = np.linspace(np.log10(SFR_v[0]), np.log10(SFR_v[-1]), 100, endpoint=True)
        CS3 = plt.contourf(Z, levels, cmap=cm)
        plt.clf()
        fig, ax = plt.subplots(1,1, figsize=(7,5))
        SFR_colormap = (SFR_v)#np.log10(np.logspace(np.log10(SFR[0]), np.log10(SFR[-1]), 10, endpoint=True))
        currentColors = [cm(1.*i/num_colors) for i in range(num_colors)]
        currentColor = iter(currentColors)
        for i,gwIMF in enumerate(gwIMF_bySFR_eval):
            ax.loglog(star_v,gwIMF, linewidth=3, c=next(currentColor))
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad="2%")
        metallicity = np.log10(metal_mass_fraction/self.solar_metallicity)
        ax.set_title(f'[Z] = {metallicity:.2f}', fontsize=15)
        ax.set_ylabel(r'$\xi_{gwIMF}$'+f' [#/{Msun}]', fontsize=15)
        ax.set_xlabel(r'stellar mass [%s]' %(Msun), fontsize=15)
        #ax.set_ylim(1e-1,1e5)
        ax.tick_params(width=2)
        cbar = fig.colorbar(CS3, cmap=cm, cax=cax, format="%.2f", ticks=ticker.MultipleLocator(1)).set_label(label=r'$\log_{10}({\rm SFR})$',size=15)
        fig.tight_layout()
        plt.savefig(f'figs/gwIMF_plots_Z{metallicity:.2f}.pdf', bbox_inches='tight')
        #plt.show(block=False)
        return None
    
    def IMF_plot(self, Mstar_v, IMF_v, Mtot, metallicity, SFR):
        from matplotlib import pyplot as plt
        import matplotlib.ticker as ticker
        Msun = r'$M_{\odot}$'
        fig, ax = plt.subplots(1,1, figsize=(6,4))
        ax.loglog(Mstar_v, IMF_v, linewidth=3, color='navy')
        ax.scatter(Mstar_v, IMF_v, linewidth=3, color='navy',s=1)
        ax.set_ylabel(r'$\xi_{\star}={\rm d} N_{\star}/{\rm d} m$'
                      +f' [#/{Msun}]', fontsize=15)
        ax.set_xlabel(r'$M_{\rm star}$ [%s]'%(Msun), fontsize=15)
        plt.title(r'$M_{\rm ecl}$ = %.2e [%s],$\quad$ [Z] = %.2f' 
                  %(Mtot, Msun, metallicity), fontsize=15)
        #ax.set_ylim(1e-8,1)
        plt.yticks(fontsize=15)
        plt.xticks(fontsize=15)
        ax.tick_params(width=2)
        fig.tight_layout()
        plt.savefig(f'figs/IMF_plot_Mecl{Mtot:.2e}_Z{metallicity:.2f}'+
                      f'_SFR{SFR:.2e}.pdf', bbox_inches='tight')
        #plt.show(block=False)
        
    def IMF_plots(self, mstar_v, IMF_v_list, Mecl_v, k_idx, massfrac):
        from matplotlib import pyplot as plt
        import colorcet as cc
        import matplotlib.ticker as ticker
        from mpl_toolkits.axes_grid1 import make_axes_locatable
        Msun = r'$M_{\odot}$'
        cm = cc.cm.CET_L20
        eff_Mecl_v = Mecl_v[k_idx]
        eff_IMF_v_list = np.array(IMF_v_list)[k_idx]
        num_colors = len(eff_Mecl_v)
        Z = [[0,0],[0,0]]
        #levels = np.linspace(np.log10(SFR_v[0]), np.log10(SFR_v[-1]), num_colors, endpoint=True)
        levels = np.linspace(np.log10(eff_Mecl_v[0]), np.log10(eff_Mecl_v[-1]), num_colors, endpoint=True)
        CS3 = plt.contourf(Z, levels, cmap=cm)
        plt.clf()
        fig, ax = plt.subplots(1,1, figsize=(7,5))
        currentColors = [cm(1.*i/num_colors) for i in range(num_colors)]
        currentColor = iter(currentColors)
        for i, IMF in enumerate(eff_IMF_v_list):
            ax.loglog(mstar_v, IMF, linewidth=3, c=next(currentColor))
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad="2%")
        ax.set_ylabel(r'$\xi_{\star}={\rm d} N_{\star}/{\rm d} m$'
                      +f' [#/{Msun}]', fontsize=15)
        ax.set_xlabel(r'$M_{\rm star}$ [%s]' %(Msun), fontsize=15)
        ax.set_ylim(1e-2,1e10)
        Z = np.log10(massfrac/0.0134)
        ax.set_title(r"[Z] = %.2f"%(Z), fontsize=15)
        ax.tick_params(width=2)
        cbar = fig.colorbar(CS3, cmap=cm, cax=cax, format="%.2f", ticks=ticker.MultipleLocator(1)).set_label(label=r'$\log_{10}(M_{\rm ecl})$'+f' [({Msun})]',size=15)
        fig.tight_layout()
        plt.savefig(f'figs/IMF_plots_Z{Z:.2f}.pdf', bbox_inches='tight')
        #plt.show(block=False)
        return None
        
    def IMF_3D_plot(self, m_v, M_ecl_v, sIMF_func):
        from mpl_toolkits import mplot3d
        import matplotlib.pyplot as plt
        
        def z_func(m,M):
            return np.reshape([[sIMF_func[i](m_j) for m_j in m] for i,val in enumerate(M)], (len(m), len(M)))
        
        def resh(x):
            return np.reshape(list(x) * len(x), (len(x),len(x)))
        
        fig = plt.figure(figsize=(10,8))
        # syntax for 3-D projection
        ax = plt.axes(projection ='3d')
        m = resh(m_v)
        M = resh(M_ecl_v).T
        xi = z_func(m_v, M_ecl_v)
        
        # plotting
        #ax.plot3D(x, y, z, 'green')
        ax.plot_surface(np.log10(m), np.log10(M), np.ma.log10(xi), cmap ='plasma', linewidth=0.25)
        ax.set_xlabel(r'stellar mass $m_{\star}$ [$\log_{10}(M_{\odot})$]', fontsize=15)
        ax.set_ylabel(r'E. cluster mass $M_{\rm ecl}$ [$\log_{10}(M_{\odot})$]', fontsize=15)
        ax.set_zlabel(r'$\xi_{\star}={\rm d}N_{\star}/ {\rm d} m$'
                      +f' [#/{Msun}]', fontsize=15)
        ax.set_title(r'stellar IMF $\xi_{\star}(m_{\star},M_{\rm ecl},Z)$'+f' [#/{Msun}]', fontsize=17)
        fig.tight_layout()
        plt.show(block=False)
        #plt.savefig(f'figs/IMF_plot_3D.pdf', bbox_inches='tight')
    
    def IGIMF_3D_plot(self, df, SFR_v, metal_mass_fraction_v, mstar_v, 
                      by_v='SFR', col_ax_idx=10, azim_rot=-120, elev_rot=20):
        '''
        by_v can be "SFR" or "metal_mass_fraction"
        '''
        from mpl_toolkits import mplot3d
        import matplotlib.pyplot as plt
        
        Msun = r'$M_{\odot}$'
        if by_v == 'SFR':
            y_ax = SFR_v
            color_ax = metal_mass_fraction_v
            title = '[Z]'
            metallicity_val = np.log10(color_ax[col_ax_idx]/0.0134)
            units = f'[{Msun}/yr]'
        elif by_v == 'metal_mass_fraction':
            y_ax = metal_mass_fraction_v
            color_ax = SFR_v
            title = 'SFR'
            units = ''
        else:
            raise ValueError("set by_v either to 'SFR' or 'metal_mass_fraction'. ")
        
        fig = plt.figure(figsize=(10,8))
        ax = plt.axes(projection ='3d')
        x = np.outer(mstar_v, np.ones(len(y_ax)))
        y = np.outer(y_ax, np.ones(len(mstar_v))).T
        xi = np.reshape([df.loc[((df['mass_star']==ival) & (df[by_v]==jval) 
            & (df['metal_mass_fraction']==metal_mass_fraction_v[col_ax_idx])
            )]['IGIMF'].to_numpy()[0] for i,ival in enumerate(mstar_v) 
            for j,jval in enumerate(y_ax)], (len(mstar_v), len(y_ax)))
        
        ## Setting a mask to exclude zero values
        #xi_mask = np.ma.masked_where(np.isnan(xi), xi)
        #xi_masked = xi.copy()
        #xi_masked[np.isnan(xi)] = -0.
        
        #ax.plot_surface(np.log10(x), np.log10(y), np.log10(xi_masked), cmap ='plasma', linewidth=0.25)
        ax.plot_surface(np.log10(x[:47,:47]), np.log10(y[:47,:47]), np.ma.log10(xi[:47,:47]), cmap ='plasma', linewidth=0.25)
        #ax.plot_surface(np.log10(x), np.log10(y), np.log10(xi_masked), cmap ='plasma', linewidth=0.25)
        ax.set_xlabel(r'stellar mass $m_{\star}$ [$\log_{10}(M_{\odot})$]', fontsize=15)
        ax.set_ylabel(f'{by_v}  {units}', fontsize=15)
        ax.set_zlabel(r'$\xi_{\rm IGIMF}={\rm d}N_{\star}/ {\rm d} m$'+
                      f'['+r'$\log_{10}({\rm #}/M_{\odot})$'+f']', fontsize=15)
        ax.set_title(f'{title} {metallicity_val:.2f}', fontsize=17)
        ax.azim = azim_rot
        ax.elev = elev_rot
        fig.tight_layout()
        plt.show(block=False)
        plt.savefig(f'figs/IGIMF_plot_3D.pdf', bbox_inches='tight')
     
    def IGIMF_3Dlines_plot(self, df, SFR_v, metal_mass_fraction_v, mstar_v):
        #from mpl_toolkits import mplot3d
        import matplotlib.pyplot as plt
        import matplotlib as mpl
        from mpl_toolkits.mplot3d import Axes3D
        import colorcet as cc
        import itertools
        mpl.rcParams['legend.fontsize'] = 10
        M = r'$M_{\odot}$'
        cm = cc.cm.CET_L8
        num_colors=len(metal_mass_fraction_v)
        currentColors = [cm(1.*i/num_colors) for i in range(num_colors)]
        currentColor = itertools.cycle(currentColors)
        fig = plt.figure(figsize=(10,8))
        ax = fig.add_subplot(projection ='3d')       
        for m in metal_mass_fraction_v: 
            for s in SFR_v:
                grid_sel = df.loc[(df['SFR']==s) & (df['metal_mass_fraction']==m)]
                ax.loglog(grid_sel['mass_star'], grid_sel['SFR'], grid_sel['IGIMF'], color=next(currentColor))
        fig.tight_layout()
        plt.show(block=False)
        plt.savefig(f'figs/IGIMF_plot_3Dlines.pdf', bbox_inches='tight')
        return None
       
    def sIMF_subplot_old(self, metallicity_v, Mecl_v, mstar_v, sIMF, res=20):
        import matplotlib.pyplot as plt 
        import itertools
        import colorcet as cc
        import matplotlib.ticker as ticker
        from mpl_toolkits.axes_grid1 import make_axes_locatable
        Msun = r'$M_{\odot}$' 
        cm = cc.cm.CET_R2
        levels = np.linspace(metallicity_v[0], metallicity_v[-1], 100,
                             endpoint=True)
        CS3 = plt.contourf([[0,0],[0,0]], levels, cmap=cm)
        plt.clf()
        num_colors=len(metallicity_v)
        currentColors = [cm(1.*i/num_colors) for i in range(num_colors)]
        currentColor = itertools.cycle(currentColors)
        nrow, ncol = 3,3 #util.find_closest_prod(res)
        fig, axs = plt.subplots(nrow, ncol, figsize=(7,5))
        for i, ax in enumerate(axs.flat):
            for j, Z in enumerate(metallicity_v):
                ax.annotate(r'$M_{\rm ecl}=$%.2e'%(Mecl_v[i]), xy=(0.5, 0.9),
                        xycoords='axes fraction', verticalalignment='top', 
                        horizontalalignment='center', fontsize=10, alpha=1)
                ax.loglog(mstar_v, sIMF[i][j], color=next(currentColor),
                          alpha=1)
                ax.set_ylim(5e-3,1e11)
                ax.set_xlim(2e-2,5e2)
        #for nr in range(3):
        for nr in range(nrow):
            for nc in range(ncol):
                if nc != 0:
                    axs[nr,nc].set_yticklabels([])
                #if nr != 3-1:
                if nr != 4-1:
                    axs[nr,nc].set_xticklabels([])
        axs[nrow//2,0].set_ylabel(r'$\xi_{\star}={\rm d} N_{\star}/{\rm d} m$'+
                                  f' [#/{Msun}]', fontsize = 15)
        axs[nrow-1, ncol//2].set_xlabel(r'stellar mass [$M_{\odot}$]',
                                        fontsize = 15)
        #divider = make_axes_locatable(axs.flat[-1])
        plt.subplots_adjust(bottom=0., right=0.95, top=1.)
        cax = plt.axes([0.85, 0.2, 0.025, 0.7])
        cbar = plt.colorbar(CS3, cmap=cm, cax=cax, format="%.2f", 
                            ticks=ticker.MultipleLocator(1)).set_label(
                                label=r'[Z]',size=15)
        fig.tight_layout(rect=[0,0,0.85,1])
        fig.subplots_adjust(wspace=0., hspace=0.)
        fig.savefig('figs/stellarIMF_subplots_Zcolorbar.pdf')



    def mw_sIMF_subplot_old(self, metallicity_v, Mecl_v, mstar_v, mw_sIMF, res=20):
        import matplotlib.pyplot as plt 
        import itertools
        import colorcet as cc
        import matplotlib.ticker as ticker
        from mpl_toolkits.axes_grid1 import make_axes_locatable
        Msun = r'$M_{\odot}$' 
        cm = cc.cm.CET_R3
        levels = np.linspace(metallicity_v[0], metallicity_v[-1], 100,
                             endpoint=True)
        CS3 = plt.contourf([[0,0],[0,0]], levels, cmap=cm)
        plt.clf()
        num_colors=len(metallicity_v)
        currentColors = [cm(1.*i/num_colors) for i in range(num_colors)]
        currentColor = itertools.cycle(currentColors)
        nrow, ncol = 3,3 #util.find_closest_prod(res)
        fig, axs = plt.subplots(nrow, ncol, figsize=(7,5))
        for i, ax in enumerate(axs.flat):
            for j, Z in enumerate(metallicity_v):
                ax.annotate(r'$M_{\rm ecl}=$%.2e'%(Mecl_v[i]), xy=(0.5, 0.9),
                        xycoords='axes fraction', verticalalignment='top', 
                        horizontalalignment='center', fontsize=10, alpha=1)
                ax.loglog(mstar_v, np.divide(mw_sIMF[i][j], Mecl_v[i]), color=next(currentColor),
                          alpha=0.8)
                ax.set_ylim(1e-8,1e3)
                ax.set_xlim(2e-2,5e2)
        #for nr in range(3):
        for nr in range(nrow):
            for nc in range(ncol):
                if nc != 0:
                    axs[nr,nc].set_yticklabels([])
                #if nr != 3-1:
                if nr != 4-1:
                    axs[nr,nc].set_xticklabels([])
        axs[nrow//2,0].set_ylabel(r'$m \xi_{\star}(m) / M_{\rm ecl} \propto \frac{{\rm d} N / {\rm d} \log_{10}m}{M_{\rm ecl}} \quad$ [#/$M_{\odot}$]', fontsize = 14)
        axs[nrow-1, ncol//2].set_xlabel(r'stellar mass [$M_{\odot}$]',
                                        fontsize = 15)
        #divider = make_axes_locatable(axs.flat[-1])
        plt.subplots_adjust(bottom=0., right=0.95, top=1.)
        cax = plt.axes([0.85, 0.2, 0.025, 0.7])
        cbar = plt.colorbar(CS3, cmap=cm, cax=cax, format="%.2f", 
                            ticks=ticker.MultipleLocator(1)).set_label(
                                label=r'[Z]',size=15)
        fig.tight_layout(rect=[0,0,0.85,1])
        fig.subplots_adjust(wspace=0., hspace=0.)
        fig.savefig('figs/massweighted_stellarIMF_subplots_Zcolorbar.pdf')

    def sIMF_subplot_Mecl_old(self, metallicity_v, Mecl_v, mstar_v, sIMF, res=20):
        import matplotlib.pyplot as plt 
        import itertools
        import colorcet as cc
        import matplotlib.ticker as ticker
        #from mpl_toolkits import mplot3d
        #from mpl_toolkits.mplot3d import Axes3D
        from mpl_toolkits.axes_grid1 import make_axes_locatable
        Msun = r'$M_{\odot}$'
        #cm = plt.cm.get_cmap(name='viridis')
        cm = cc.cm.CET_L20
        levels = np.linspace(np.log10(Mecl_v[0]), np.log10(Mecl_v[-1]), 100,
                             endpoint=True)
        CS3 = plt.contourf([[0,0],[0,0]], levels, cmap=cm)
        plt.clf()
        num_colors=len(Mecl_v)
        currentColors = [cm(1.*i/num_colors) for i in range(num_colors)]
        currentColor = itertools.cycle(currentColors)
        nrow, ncol = 3, 3 #util.find_closest_prod(res)
        #fig, axs = plt.subplots(3, 3, figsize=(8,6))
        fig, axs = plt.subplots(nrow, ncol, figsize=(7,5))
        for i, ax in enumerate(axs.flat):
            for j, M in enumerate(Mecl_v):
                ax.annotate(r'$[Z]=$%.2f'%(metallicity_v[i]), xy=(0.5, 0.9), 
                        xycoords='axes fraction', verticalalignment='top', 
                        horizontalalignment='center', fontsize=10, alpha=.1)
                ax.loglog(mstar_v, sIMF[j][i], color=next(currentColor))
                ax.set_ylim(5e-3,1e11)
                ax.set_xlim(2e-2,5e2)
        #for nr in range(3):
        for nr in range(nrow):
            #for nc in range(3):
            for nc in range(ncol):
                if nc != 0:
                    axs[nr,nc].set_yticklabels([])
                #if nr != 3-1:
                if nr != nrow-1:
                    axs[nr,nc].set_xticklabels([])
        #axs[1,0].set_ylabel(r'$\xi_{stellar}$', fontsize = 15)
        #axs[2,1].set_xlabel(r'stellar mass [$M_{\odot}$]', fontsize = 15)
        axs[nrow//2,0].set_ylabel(r'$\xi_{\star}={\rm d} N_{\star}/{\rm d} m$'
                                  +f' [#/{Msun}]', fontsize = 15)
        axs[nrow-1,ncol//2].set_xlabel(r'stellar mass [$M_{\odot}$]', 
                                       fontsize = 15)
        #divider = make_axes_locatable(axs.flat[-1])
        plt.subplots_adjust(bottom=0., right=0.95, top=1.)
        cax = plt.axes([0.85, 0.2, 0.025, 0.7])
        cbar = plt.colorbar(CS3, cmap=cm, cax=cax, format="%.2f", 
                            ticks=ticker.MultipleLocator(1)).set_label(
                                label=r'$\log_{10}(M_{\rm ecl})$'+f' ['+
                                r'$\log_{10}$'+f'({Msun})]',size=15)
        fig.tight_layout(rect=[0,0,0.85,1])
        fig.subplots_adjust(wspace=0., hspace=0.)
        fig.savefig('figs/stellarIMF_subplots_Meclcolorbar.pdf')

    def sIMF_subplot(self, metallicity_v, Mecl_v, mstar_v, sIMF, res=20):
        import matplotlib.pyplot as plt 
        import itertools
        import colorcet as cc
        import matplotlib.ticker as ticker
        from mpl_toolkits.axes_grid1 import make_axes_locatable
        Msun = r'$M_{\odot}$' 
        cm = cc.cm.CET_R2
        levels = np.linspace(metallicity_v[0], metallicity_v[-1], 100,
                             endpoint=True)
        CS3 = plt.contourf([[0,0],[0,0]], levels, cmap=cm)
        plt.clf()
        num_colors=len(metallicity_v)
        currentColors = reversed(list([cm(1.*i/num_colors) for i in range(num_colors)]))
        currentColor = itertools.cycle(currentColors)
        nrow, ncol = 2,2 #3,3 #util.find_closest_prod(res)
        fig, axs = plt.subplots(nrow, ncol, figsize=(7,5))
        for i, ax in enumerate(axs.flat):
            for j, Z in reversed(list(enumerate(metallicity_v))):
                ax.annotate(r'$M_{\rm ecl}=$%.2e'%(Mecl_v[i]), xy=(0.5, 0.9),
                        xycoords='axes fraction', verticalalignment='top', 
                        horizontalalignment='center', fontsize=10, alpha=1)
                ax.loglog(mstar_v, sIMF[i][j], color=next(currentColor),
                          alpha=1)
                for shift in np.arange(-5,20):
                    ax.loglog(mstar_v, util.Kroupa01(mstar_v)*np.power(10.,shift), color='grey', linewidth=0.2, linestyle='--', alpha=0.1)
                ax.set_ylim(5e-3,1e11)
                ax.set_xlim(6e-2,1.6e2)
        #for nr in range(3):
        for nr in range(nrow):
            #for nc in range(3):
            for nc in range(ncol):
                if nc != 0:
                    axs[nr,nc].set_yticklabels([])
                #if nr != 3-1:
                if nr != nrow-1:
                    axs[nr,nc].set_xticklabels([])
        axs[nrow//2,0].set_ylabel(r'$\xi_{\star}={\rm d} N_{\star}/{\rm d} m$'+
                                  f' [#/{Msun}]', fontsize = 15)
        axs[nrow-1, ncol//2].set_xlabel(r'stellar mass [$M_{\odot}$]',
                                        fontsize = 15)
        axs[nrow//2,0].yaxis.set_label_coords(-.15, 1)
        axs[nrow-1, ncol//2].xaxis.set_label_coords(0., -.15)
        #divider = make_axes_locatable(axs.flat[-1])
        plt.subplots_adjust(bottom=0., right=0.95, top=1.)
        cax = plt.axes([0.85, 0.2, 0.025, 0.7])
        cbar = plt.colorbar(CS3, cmap=cm, cax=cax, format="%.2f", 
                            ticks=ticker.MultipleLocator(1)).set_label(
                                label=r'[Z]',size=15)
        fig.tight_layout(rect=[0,0,0.85,1])
        fig.subplots_adjust(wspace=0., hspace=0.)
        fig.savefig('figs/stellarIMF_subplots_Zcolorbar.pdf')



    def mw_sIMF_subplot(self, metallicity_v, Mecl_v, mstar_v, mw_sIMF, res=20):
        import matplotlib.pyplot as plt 
        import itertools
        import colorcet as cc
        import matplotlib.ticker as ticker
        from mpl_toolkits.axes_grid1 import make_axes_locatable
        Msun = r'$M_{\odot}$' 
        cm = cc.cm.CET_R3
        levels = np.linspace(metallicity_v[0], metallicity_v[-1], 100,
                             endpoint=True)
        CS3 = plt.contourf([[0,0],[0,0]], levels, cmap=cm)
        plt.clf()
        num_colors=len(metallicity_v)
        currentColors = reversed(list([cm(1.*i/num_colors) for i in range(num_colors)]))
        currentColor = itertools.cycle(currentColors)
        nrow, ncol = 2,2 #util.find_closest_prod(res)
        fig, axs = plt.subplots(nrow, ncol, figsize=(7,5))
        for i, ax in enumerate(axs.flat):
            for j, Z in reversed(list(enumerate(metallicity_v))):
                ax.annotate(r'$M_{\rm ecl}=$%.2e'%(Mecl_v[i]), xy=(0.5, 0.9),
                        xycoords='axes fraction', verticalalignment='top', 
                        horizontalalignment='center', fontsize=10, alpha=1)
                ax.loglog(mstar_v, np.divide(mw_sIMF[i][j], Mecl_v[i]), color=next(currentColor),
                          alpha=0.8)
                for shift in np.arange(-5,20):
                    ax.loglog(mstar_v, np.divide(mstar_v*util.Kroupa01(mstar_v)*np.power(10.,shift),Mecl_v[i]), color='grey', linewidth=0.2, linestyle='--', alpha=0.1)
                #ax.set_xlim(2e-2,5e2)
                #ax.set_ylim(5e-3,1e11)
                ax.set_xlim(6e-2,1.6e2)
                ax.set_ylim(1e-8,1e3)
        #for nr in range(3):
        for nr in range(nrow):
            #for nc in range(3):
            for nc in range(ncol):
                if nc != 0:
                    axs[nr,nc].set_yticklabels([])
                #if nr != 3-1:
                if nr != nrow-1:
                    axs[nr,nc].set_xticklabels([])
        axs[nrow//2,0].set_ylabel(r'$m \xi_{\star}(m) / M_{\rm ecl} \propto \frac{{\rm d} N / {\rm d} \log_{10}m}{M_{\rm ecl}} \quad$ [#/$M_{\odot}$]', fontsize = 14)
        axs[nrow-1, ncol//2].set_xlabel(r'stellar mass [$M_{\odot}$]',
                                        fontsize = 15)
        axs[nrow//2,0].yaxis.set_label_coords(-.2, 1)
        axs[nrow-1, ncol//2].xaxis.set_label_coords(0., -.15)
        #divider = make_axes_locatable(axs.flat[-1])
        plt.subplots_adjust(bottom=0., right=0.95, top=1.)
        cax = plt.axes([0.85, 0.2, 0.025, 0.7])
        cbar = plt.colorbar(CS3, cmap=cm, cax=cax, format="%.2f", 
                            ticks=ticker.MultipleLocator(1)).set_label(
                                label=r'[Z]',size=15)
        fig.tight_layout(rect=[0,0,0.85,1])
        fig.subplots_adjust(wspace=0., hspace=0.)
        fig.savefig('figs/massweighted_stellarIMF_subplots_Zcolorbar.pdf')

    def sIMF_subplot_Mecl(self, metallicity_v, Mecl_v, mstar_v, sIMF, res=20):
        import matplotlib.pyplot as plt 
        import itertools
        import colorcet as cc
        import matplotlib.ticker as ticker
        #from mpl_toolkits import mplot3d
        #from mpl_toolkits.mplot3d import Axes3D
        from mpl_toolkits.axes_grid1 import make_axes_locatable
        Msun = r'$M_{\odot}$'
        #cm = plt.cm.get_cmap(name='viridis')
        cm = cc.cm.CET_I1
        levels = np.linspace(np.log10(Mecl_v[0]), np.log10(Mecl_v[-1]), 100,
                             endpoint=True)
        CS3 = plt.contourf([[0,0],[0,0]], levels, cmap=cm)
        plt.clf()
        num_colors=len(Mecl_v)
        currentColors = [cm(1.*i/num_colors) for i in range(num_colors)]
        currentColor = itertools.cycle(currentColors)
        nrow, ncol = 2,2 #util.find_closest_prod(res)
        #fig, axs = plt.subplots(3, 3, figsize=(8,6))
        fig, axs = plt.subplots(nrow, ncol, figsize=(7,5))
        for i, ax in enumerate(axs.flat):
            for j, M in enumerate(Mecl_v):
                ax.annotate(r'$[Z]=$%.2f'%(metallicity_v[i]), xy=(0.5, 0.9), 
                        xycoords='axes fraction', verticalalignment='top', 
                        horizontalalignment='center', fontsize=10, alpha=.1)
                ax.loglog(mstar_v, sIMF[j][i], color=next(currentColor))
                for shift in np.arange(-5,20):
                    ax.loglog(mstar_v, util.Kroupa01(mstar_v)*np.power(10.,shift), color='grey', linewidth=0.2, linestyle='--', alpha=0.1)
                ax.set_ylim(5e-3,1e11)
                ax.set_xlim(6e-2,1.6e2)
        #for nr in range(3):
        for nr in range(nrow):
            #for nc in range(3):
            for nc in range(ncol):
                if nc != 0:
                    axs[nr,nc].set_yticklabels([])
                #if nr != 3-1:
                if nr != nrow-1:
                    axs[nr,nc].set_xticklabels([])
        #axs[1,0].set_ylabel(r'$\xi_{stellar}$', fontsize = 15)
        #axs[2,1].set_xlabel(r'stellar mass [$M_{\odot}$]', fontsize = 15)
        axs[nrow//2,0].set_ylabel(r'$\xi_{\star}={\rm d} N_{\star}/{\rm d} m$'
                                  +f' [#/{Msun}]', fontsize = 15)
        axs[nrow-1,ncol//2].set_xlabel(r'stellar mass [$M_{\odot}$]', 
                                       fontsize = 15)
        axs[nrow//2,0].yaxis.set_label_coords(-.15, 1)
        axs[nrow-1, ncol//2].xaxis.set_label_coords(0., -.15)
        #divider = make_axes_locatable(axs.flat[-1])
        plt.subplots_adjust(bottom=0., right=0.95, top=1.)
        cax = plt.axes([0.85, 0.2, 0.025, 0.7])
        cbar = plt.colorbar(CS3, cmap=cm, cax=cax, format="%.2f", 
                            ticks=ticker.MultipleLocator(1)).set_label(
                                label=r'$\log_{10}(M_{\rm ecl})$'+f' ['+
                                r'$\log_{10}$'+f'({Msun})]',size=15)
        fig.tight_layout(rect=[0,0,0.85,1])
        fig.subplots_adjust(wspace=0., hspace=0.)
        fig.savefig('figs/stellarIMF_subplots_Meclcolorbar.pdf')


    def sIMF_subplot_SFR(self, metallicity_v, SFR_v, mstar_v, sIMF, res=20):
        import matplotlib.pyplot as plt 
        import itertools
        import colorcet as cc
        import matplotlib.ticker as ticker
        #from mpl_toolkits import mplot3d
        #from mpl_toolkits.mplot3d import Axes3D
        from mpl_toolkits.axes_grid1 import make_axes_locatable
        Msun = r'$M_{\odot}$'
        #cm = plt.cm.get_cmap(name='viridis')
        cm = cc.cm.CET_I1
        levels = np.linspace(np.log10(SFR_v[0]), np.log10(SFR_v[-1]), 100,
                             endpoint=True)
        CS3 = plt.contourf([[0,0],[0,0]], levels, cmap=cm)
        plt.clf()
        num_colors=len(SFR_v)
        currentColors = [cm(1.*i/num_colors) for i in range(num_colors)]
        currentColor = itertools.cycle(currentColors)
        nrow, ncol = 2,2 #util.find_closest_prod(res)
        #fig, axs = plt.subplots(3, 3, figsize=(8,6))
        fig, axs = plt.subplots(nrow, ncol, figsize=(7,5))
        for i, ax in enumerate(axs.flat):
            for j, M in enumerate(SFR_v):
                ax.annotate(r'$[Z]=$%.2f'%(metallicity_v[i]), xy=(0.5, 0.9), 
                        xycoords='axes fraction', verticalalignment='top', 
                        horizontalalignment='center', fontsize=10, alpha=.1)
                ax.loglog(mstar_v, sIMF[j][i], color=next(currentColor))
                for shift in np.arange(-5,20):
                    ax.loglog(mstar_v, util.Kroupa01(mstar_v)*np.power(10.,shift), color='grey', linewidth=0.2, linestyle='--', alpha=0.1)
                ax.set_ylim(5e-3,1e11)
                ax.set_xlim(6e-2,1.6e2)
        #for nr in range(3):
        for nr in range(nrow):
            #for nc in range(3):
            for nc in range(ncol):
                if nc != 0:
                    axs[nr,nc].set_yticklabels([])
                #if nr != 3-1:
                if nr != nrow-1:
                    axs[nr,nc].set_xticklabels([])
        #axs[1,0].set_ylabel(r'$\xi_{stellar}$', fontsize = 15)
        #axs[2,1].set_xlabel(r'stellar mass [$M_{\odot}$]', fontsize = 15)
        axs[nrow//2,0].set_ylabel(r'$\xi_{\star}={\rm d} N_{\star}/{\rm d} m$'
                                  +f' [#/{Msun}]', fontsize = 15)
        axs[nrow-1,ncol//2].set_xlabel(r'stellar mass [$M_{\odot}$]', 
                                       fontsize = 15)
        axs[nrow//2,0].yaxis.set_label_coords(-.15, 1)
        axs[nrow-1, ncol//2].xaxis.set_label_coords(0., -.15)
        #divider = make_axes_locatable(axs.flat[-1])
        plt.subplots_adjust(bottom=0., right=0.95, top=1.)
        cax = plt.axes([0.85, 0.2, 0.025, 0.7])
        cbar = plt.colorbar(CS3, cmap=cm, cax=cax, format="%.2f", 
                            ticks=ticker.MultipleLocator(1)).set_label(
                                label=r'$\log_{10}$(SFR)'+f' ['+
                                r'$\log_{10}$'+f'({Msun}/yr)]',size=15)
        fig.tight_layout(rect=[0,0,0.85,1])
        fig.subplots_adjust(wspace=0., hspace=0.)
        fig.savefig('figs/stellarIMF_subplots_SFRcolorbar.pdf')

    def alpha3_plot(self, alpha3_v, x_v, parameter_space):
        from matplotlib import pyplot as plt
        import matplotlib.ticker as ticker
        import colorcet as cc
        fig,ax = plt.subplots(figsize=(6,4))
        x,y = np.meshgrid(np.log10(parameter_space['rho_cl']), parameter_space['[Z]'])
        z = np.array(alpha3_v).reshape(resolution,resolution).T
        cax = ax.contourf(x, y, z, resolution, cmap=cc.cm.CET_L6)
        plt.xlabel(r'$\log_{10}(\rho_{cl})$ [$\log_{10}(M_{\odot}/{\rm pc}^3)$]', fontsize=15)
        plt.ylabel(r'[$Z$]', fontsize=15)
        cbar = fig.colorbar(cax)
        cbar.set_label(r'$\alpha_3$', fontsize=15)
        plt.tight_layout()
        plt.savefig('figs/alpha3plot.pdf')

    def Cook23_plot(self):
        from matplotlib import pyplot as plt
        import matplotlib.ticker as ticker
        import pandas as pd
        import numpy as np
        Cook23 = pd.read_csv('data/Cook23.dat', comment='#', sep='&')
        Cook23bin = pd.read_csv('data/Cook23bin.dat', comment='#', sep='&')
        Cook23bin10 = Cook23bin.loc[Cook23bin['SFR-Method']=='1-10Myr']
        Cook23binHalpha = Cook23bin.loc[Cook23bin['SFR-Method']=='Halpha']
        Dinnbier22 = pd.read_csv('data/Dinnbier22.dat', comment='#', sep=',')
        D22low = Dinnbier22.iloc[:3]
        D22high = Dinnbier22.iloc[3:]
        fig, ax = plt.subplots(1,1, figsize=(7,5))
        ax.axhline(y=100, xmin=-4, xmax=1, linewidth=4, color='purple', label='IGIMF at birth')
        ax.semilogy(D22low['logSFR'], np.power(10, D22low['Gamma'])*100, linestyle='--', color='purple', linewidth=2)
        ax.semilogy(D22high['logSFR'], np.power(10, D22high['Gamma'])*100, linestyle='--', color='purple', linewidth=2)
        ax.fill_between(D22high['logSFR'], np.power(10, D22high['Gamma'])*100, np.power(10, D22low['Gamma'])*100, where=(np.power(10, D22low['Gamma'].to_numpy())*100<np.power(10, D22high['Gamma'].to_numpy())*100),  alpha=0.1, color='purple', label=r'DKA22 $<10$ Myr')
        ax.errorbar(Cook23binHalpha['sfrsig-bin'], Cook23binHalpha['Gamma'], xerr=Cook23binHalpha['sfrsig-u'], yerr=Cook23binHalpha['Gamma-u'], fmt='o', color='red', ecolor='red', elinewidth=3, capsize=0, label=r'C23 H$_{\alpha}$ $<10$ Myr', marker='s', alpha=0.8)
        ax.errorbar(Cook23bin10['sfrsig-bin'], Cook23bin10['Gamma'], xerr=Cook23bin10['sfrsig-u'], yerr=Cook23bin10['Gamma-u'], fmt='o', color='black', ecolor='black', elinewidth=3, capsize=0,label=r'C23 res $<10$ Myr', marker='s', alpha=0.8)
        #ax.errorbar(Cook23['logSFRsig'], Cook23['Gamma'], yerr=Cook23['Gamma-u'], fmt='o', color='blue', ecolor='blue', elinewidth=1, capsize=0,label=r'C23 lit', marker='o', alpha=0.4)
        ax.set_xlim(-3.7, 0.5)
        ax.set_ylim(2e-2, 2e2)
        ax.legend(loc='lower right', fontsize=12, frameon=True)
        ax.set_ylabel(r'$\Gamma$ (%)', fontsize=15)
        ax.set_xlabel(r'$\log(\Sigma_{\rm SFR})$ ($M_{\star} yr^{-1} kpc^{-2}$)', fontsize=15)
        fig.tight_layout()
        plt.savefig('figs/Cook23.pdf', bbox_inches='tight')
        #plt.show(block=False)

    def Fig11_plot(self):
        from matplotlib import pyplot as plt
        import matplotlib.ticker as ticker
        CMOl = np.loadtxt('../data/Capuzzo-dolcetta17CMOl.csv', delimiter=',')
        CMOu = np.loadtxt('../data/Capuzzo-dolcetta17CMOu.csv', delimiter=',')
        SMBH = np.loadtxt('../data/Capuzzo-dolcetta17BH.csv', delimiter=',')
        fig, ax = plt.subplots(1,1, figsize=(7,5))
        #ax.loglog(time, DTD_SNIa, color='blue', label='SNIa')
        #ax.legend(loc='best', frameon=False, fontsize=13)
        ax.scatter(CMOl[:,0], CMOl[:,1], color='red', marker='s', alpha=.7)
        ax.scatter(CMOu[:,0], CMOu[:,1], color='magenta', marker='^', alpha=.7)
        ax.scatter(SMBH[:,0], SMBH[:,1], color='black', marker='o', alpha=.7)
        ax.set_ylabel(r'$\log_{10}(M_{\rm CMO}/M_{\odot})$', fontsize=15)
        ax.set_xlabel(r'$\log_{10}(M_{\rm pgal}/M_{\odot})$', fontsize=15)
        
        ax.xaxis.set_minor_locator(ticker.MultipleLocator(base=.5))
        ax.tick_params(width=1, length=10, axis='x', which='minor', 
                       bottom=True, top=True, direction='in')
        ax.xaxis.set_major_locator(ticker.MultipleLocator(base=2))
        ax.tick_params(width=2, length=15, axis='x', which='major', 
                       bottom=True, top=True, direction='in')
        ax.xaxis.set_minor_locator(ticker.MultipleLocator(base=.5))
        ax.tick_params(width=1, length=10, axis='x', which='minor', 
                       bottom=True, top=True, direction='in')
        ax.xaxis.set_major_locator(ticker.MultipleLocator(base=2))
        ax.tick_params(width=2, length=15, axis='x', which='major', 
                       bottom=True, top=True, direction='in')
        plt.yticks(fontsize=15)
        plt.xticks(fontsize=15)
        ax.tick_params(width=2)
        ax.set_ylim(0,11.5)
        ax.set_xlim(6, 13.7)
        fig.tight_layout()
        plt.savefig('figs/Fig11.pdf', bbox_inches='tight')
        #plt.show(block=False)
              
    def k_Z_plot(self, Z_massfrac_v, k_IMF_Z_list, m_max_Z_list, Mecl_v,
                 m_star_max=150):
        from matplotlib import pyplot as plt
        import matplotlib.ticker as ticker
        from mpl_toolkits.axes_grid1 import make_axes_locatable
        cm = plt.cm.get_cmap(name='plasma')
        cm2 = plt.cm.get_cmap(name='plasma')
        num_colors = len(Z_massfrac_v)
        Z = [[0,0],[0,0]]
        #levels = np.linspace(np.log10(SFR_v[0]), np.log10(SFR_v[-1]), num_colors, endpoint=True)
        levels = np.linspace(np.log10(Z_massfrac_v[0]/0.0142), 
                             np.log10(Z_massfrac_v[-1]/0.0142), 
                             100, endpoint=True)
        CS3 = plt.contourf(Z, levels, cmap=cm)
        plt.clf()
        SFR_colormap = (Z_massfrac_v) #np.log10(np.logspace(np.log10(SFR[0]), np.log10(SFR[-1]), 10, endpoint=True))
        currentColors = [cm(1.*i/num_colors) for i in range(num_colors)]
        currentColor = iter(currentColors)
        currentColors2 = [cm2(1.*i/num_colors) for i in range(num_colors)]
        currentColor2 = iter(currentColors2)
        Msun = r'$M_{\odot}$'
        fig, ax = plt.subplots(2,1, figsize=(5,7))
        #ax2 = ax.twinx()
        #ax.plot(np.log10(Z_massfrac_v)-0.0142, [alpha1_Z_list[i][0] for i in range(len(SFR_v))], linewidth=3, color='magenta')
        #ax.scatter(np.log10(Z_massfrac_v)-0.0142, [alpha1_Z_list[i][0] for i in range(len(SFR_v))], linewidth=3, color='magenta')
        for i,Z in enumerate(Z_massfrac_v):
            color = next(currentColor)
            ax[1].semilogx(Mecl_v, np.log10(k_IMF_Z_list[i]), linewidth=3, color=color, alpha=0.4)
            #ax.plot((Mecl_v), (k_IMF_Z_list[i]), linewidth=3, color=color, alpha=0.4)
            color2 = next(currentColor2)
            ax[0].semilogx(Mecl_v, m_max_Z_list[i], linewidth=3, color=color2, alpha=0.4)
            #ax2.plot((Mecl_v), m_max_Z_list[i], linewidth=3, color=color2, alpha=0.4)
        ax[1].set_ylabel(r'$\log_{10}(k_{\rm IMF})$', fontsize=15)
        ax[1].set_xlabel(r'$\log_{10}(M_{\rm ecl})$[$M_{\odot}$]', fontsize=15)
        #ax.set_ylabel(r'$k_{\rm IMF}$', fontsize=15)
        #ax.set_xlabel(r'$M_{\rm ecl}$[$M_{\odot}$]', fontsize=15)
        ax[0].set_ylabel(r'$m_{\rm max}$', fontsize=15)
        divider = make_axes_locatable(ax[0])
        cax = divider.append_axes("top", size="5%", pad="10%")#, pack_start=True)
        #cax = divider.new_vertical(size="5%", pad=2.6, pack_start=True)
        fig.add_axes(cax)
        cbar = fig.colorbar(CS3, cmap=cm, cax=cax, format="%.2f", ticks=ticker.MultipleLocator(1),orientation="horizontal").set_label(label=r'$[Z]$',size=15)
        cax.xaxis.set_ticks_position("top")
        cax.xaxis.set_label_position("top")
        fig.tight_layout()
        plt.savefig(f'figs/Mecl_vs_k_mmax{m_star_max}.pdf', bbox_inches='tight')
        #plt.show(block=False)


    def alpha1_Z_plot(self, Z_massfrac_v, alpha1_Z_list):
        from matplotlib import pyplot as plt
        import matplotlib.ticker as ticker
        Msun = r'$M_{\odot}$'
        fig, ax = plt.subplots(1,1, figsize=(7,5))
        #ax.plot(np.log10(Z_massfrac_v)-0.0142, [alpha1_Z_list[i][0] for i in range(len(SFR_v))], linewidth=3, color='magenta')
        #ax.scatter(np.log10(Z_massfrac_v)-0.0142, [alpha1_Z_list[i][0] for i in range(len(SFR_v))], linewidth=3, color='magenta')
        ax.plot(Z_massfrac_v- 0.0142, [alpha1_Z_list[i][0] for i in range(len(Z_massfrac_v))], linewidth=3, color='magenta', alpha=0.4)
        ax.scatter(Z_massfrac_v- 0.0142, [alpha1_Z_list[i][0] for i in range(len(Z_massfrac_v))], linewidth=3, color='magenta', alpha=0.4)
        ax.set_ylabel(r'$\alpha_1$', fontsize=15)
        ax.set_xlabel(r'[Z - $Z_{\odot}$]', fontsize=15)
        ax.axhline(1.3 - 63*0.0142, linestyle=':', color='orange')
        ax.axhline(1.3 + 63*0.0142, linestyle=':', color='orange')
        ax.axhline(1.3, linestyle='--', color='orange')
        ax.axvline(0, linestyle='--', color='orange')
        ax.plot(Z_massfrac_v - 0.0142, 1.3 + np.arctan(1.3e2*(Z_massfrac_v - 0.0142))/1.3, color='red', linewidth=3)
        #plt.title(r'SFR = %.2e [%s/yr]' %(self.SFR, Msun), fontsize=15)
        #ax.set_ylim(5e-2,1e8)
        #ax.set_xlim(1e-11,1e0)
        plt.yticks(fontsize=15)
        plt.xticks(fontsize=15)
        ax.tick_params(width=2)
        fig.tight_layout()
        plt.savefig(f'figs/alpha1_Z_plot_{name}.pdf', bbox_inches='tight')
        #plt.show(block=False)



import scipy.interpolate
from numpy import *
import pandas as pd

class FriendlyInterpolant(object):
    def __init__(self, df, tf_funs={}, xcols=None, ycol='y', name='model', plot=False, plot_mod=lambda fig: None, fig_root='./', plot_ops={}):
        """
        Args:
            df (pd.DataFrame): dataframe of features and response vector.
            tf_funs (dict): transform functions where keys are df column names potentially with '.inv' or prime appended '.prime'. 
                Response transforms are required to have both col and col+'.inv' keys. 
                If planning to take derivitive with respect to col, supply col+'.prime' key.
            xcols (list): list of feature col names. If None, use all col not equal to ycol. 
            ycol (str): column name of response vector.
            name (str): name of instance for plotting and printing
            plot (pd.DataFrame): columns should be all feature columns. Each row is a vector of derivitives. 
                plot=True is equivalent to having a single row of 0 values.  
                plot=False is equivalent to having no rows. 
            plot_mod (func): function which modifies input fig,ax before saving plot. 
                The plot is closed after __init__ as df is not retained.
            fig_root (str): output directory for figure
            plot_ops (dict): plotting options
        """
        for k,v in {'sepfl':0,'sepfr':0,'sepfb':0,'sepft':0,'scatter':True,'view_init_azim':45}.items():
            if k not in plot_ops: plot_ops[k] = v
        if isinstance(df,ndarray):
            assert df.ndim==2
            df = pd.DataFrame(df,columns=['x_{%d}'%i for i in range(df.shape[1]-1)]+['y'])
        self.ycol = ycol
        self.xcols = xcols
        assert self.ycol in df.columns
        if self.xcols is None: self.xcols = [xcol for xcol in list(df.columns) if xcol!=self.ycol]
        assert all(xcol in df.columns for xcol in self.xcols)
        self.d = len(self.xcols)
        assert df.shape[0]>1 and self.d>0
        assert self.ycol in df.columns
        self.xcols = [col for col in list(df.columns) if col!=ycol]
        self.tf_funs = tf_funs
        if self.ycol in self.tf_funs:
            assert self.ycol+'.inv' in self.tf_funs
        else:
            self.tf_funs[self.ycol] = lambda y:y
            self.tf_funs[self.ycol+'.inv'] = lambda y:y
            self.tf_funs[self.ycol+'.prime'] = lambda y:ones(len(y),dtype=float)
        for col in self.xcols:
            if col not in self.tf_funs:
                self.tf_funs[col] = lambda x:x
                self.tf_funs[col+'.prime'] = lambda x:ones(len(x),dtype=float)
        self.name = name        
        dftf = pd.DataFrame({col:self.tf_funs[col](df[col].to_numpy()) for col in list(df.columns)})
        self.descrip = df.describe()
        self.descrip_tf = dftf.describe()
        xtf = dftf[self.xcols].to_numpy()
        ytf = dftf[self.ycol].to_numpy()
        self._fit(xtf,ytf)       
        self.train_metrics = self.get_metrics(df)
        if plot is False or self.d>2: return
        if plot is True or plot is None: plot=[None]
        assert isinstance(plot,list)
        dwrts = [self._parse_dwrt(dwrt) for dwrt in plot]
        from matplotlib import pyplot
        nticks = 257
        if self.d==1:
            xcol = self.xcols[0]
            fig,ax = pyplot.subplots(nrows=len(dwrts),ncols=2,figsize=(4*2,4.5*len(dwrts)))
            ax = atleast_2d(ax)
            for j,dfv in enumerate([dftf,df]):
                xvmin,xvmax = dfv[xcol].min(),dfv[xcol].max()
                xvsep = xvmax-xvmin
                sepfl,sepfr = plot_ops['sepfl'],plot_ops['sepfr']
                xvticks = linspace(xvmin-sepfl*xvsep,xvmax+xvsep*sepfr,nticks)
                for i,dwrt in enumerate(dwrts):
                    if plot_ops['scatter'] and (dwrt==0).all(): ax[i,j].scatter(dfv[xcol],dfv[self.ycol],color='c')
                    yvhat = self.__call__(xvticks[:,None],dwrt) if j==1 else self._eval_with_grad(xvticks[:,None],dwrt)
                    ax[i,j].plot(xvticks,yvhat,color='k')
            for i in range(len(dwrts)):
                ax[i,0].set_xlabel(r'$\mathrm{%s}^\mathrm{tf}$'%xcol)
                ax[i,0].set_ylabel(r'$\mathrm{%s}^\mathrm{tf}$'%self.ycol)
                ax[i,0].set_title(r'$\mathrm{%s}_\mathrm{tf}^{%s}(%s^\mathrm{tf})$'%(self.name,str(tuple(dwrts[i])),xcol))
                ax[i,1].set_xlabel(r'$\mathrm{%s}$'%xcol)
                ax[i,1].set_ylabel(r'$\mathrm{%s}$'%self.ycol)
                ax[i,1].set_title(r'$\mathrm{%s}_{%s}(%s)$'%(self.name,str(tuple(dwrts[i])),xcol))
                for j in range(2):
                    (xmin,xmax),(ymin,ymax) = ax[i,j].get_xlim(),ax[i,j].get_ylim()
                    ax[i,j].set_aspect((xmax-xmin)/(ymax-ymin))
            fig.suptitle('%s = %s(%s)\nTransformed Domain | Original Domain'%(self.ycol,self.name,xcol))
        elif self.d==2:
            from matplotlib import cm
            x0col,x1col = self.xcols
            nrows = len(dwrts)
            fig = pyplot.figure(figsize=(5.5*4,4.5*nrows))
            for j,dfv in enumerate([dftf,df]):
                xv0min,xv0max,xv1min,xv1max = dfv[x0col].min(),dfv[x0col].max(),dfv[x1col].min(),dfv[x1col].max()
                xv0sep,xv1sep = xv0max-xv0min,xv1max-xv1min
                sepfl,sepfr,sepfb,sepft = plot_ops['sepfl'],plot_ops['sepfr'],plot_ops['sepfb'],plot_ops['sepft']
                xv0ticks = linspace(xv0min-sepfl*xv0sep,xv0max+sepfr*xv0sep,nticks)
                xv1ticks = linspace(xv1min-sepfb*xv1sep,xv1max+sepft*xv0sep,nticks)
                xv0mesh,xv1mesh = meshgrid(xv0ticks,xv1ticks)
                xvticks = vstack([xv0mesh.flatten(),xv1mesh.flatten()]).T
                for i,dwrt in enumerate(dwrts):
                    yv = self.__call__(xvticks,dwrt) if j==1 else self._eval_with_grad(xvticks,dwrt)
                    yvmesh = yv.reshape(xv0mesh.shape)
                    ax = fig.add_subplot(nrows,4,4*i+2*j+(1 if j==0 else 2))
                    contour = ax.contourf(xv0mesh,xv1mesh,yvmesh,cmap='gray',alpha=.95,vmin=yvmesh.min(),vmax=yvmesh.max(),levels=64)
                    cbar = fig.colorbar(contour,ax=None,shrink=1,aspect=5)
                    if plot_ops['scatter'] and (dwrt==0).all(): ax.scatter(dfv[x0col],dfv[x1col],color='c')
                    ax.set_xlabel(r'$\mathrm{%s}$'%x0col if j==1 else r'$\mathrm{%s}^\mathrm{tf}$'%x0col)
                    ax.set_ylabel(r'$\mathrm{%s}$'%x1col if j==1 else r'$\mathrm{%s}^\mathrm{tf}$'%x1col)
                    if j==1: ax.set_title(r'$\mathrm{%s}_{%s}(%s,%s)$'%(self.name,str(tuple(dwrts[i])),x0col,x1col))
                    elif j==0: ax.set_title(r'$\mathrm{%s}_\mathrm{tf}^{%s}(%s^\mathrm{tf},%s^\mathrm{tf})$'%(self.name,str(tuple(dwrts[i])),x0col,x1col))
                    (xmin,xmax),(ymin,ymax) = ax.get_xlim(),ax.get_ylim()
                    ax.set_aspect((xmax-xmin)/(ymax-ymin))
                    ax = fig.add_subplot(nrows,4,4*i+2*j+(2 if j==0 else 1),projection='3d')
                    ax.plot_surface(xv0mesh,xv1mesh,yvmesh,cmap='gray',alpha=.95,vmin=yvmesh.min(),vmax=yvmesh.max(),rcount=82,ccount=82)
                    ax.view_init(azim=plot_ops['view_init_azim'])
                    if plot_ops['scatter'] and (dwrt==0).all(): ax.scatter(dfv[x0col],dfv[x1col],dfv[self.ycol],color='c')
                    ax.set_xlabel(r'$\mathrm{%s}$'%x0col if j==1 else r'$\mathrm{%s}^\mathrm{tf}$'%x0col)
                    ax.set_ylabel(r'$\mathrm{%s}$'%x1col if j==1 else r'$\mathrm{%s}^\mathrm{tf}$'%x1col)
                    ax.set_zlabel(r'$\mathrm{%s}$'%self.ycol if j==1 else r'$\mathrm{%s}^\mathrm{tf}$'%self.ycol)
            fig.suptitle('%s = %s(%s,%s)\nTransformed Domain | Original Domain'%(self.ycol,self.name,x0col,x1col))
            #pyplot.subplots_adjust(left=0,bottom=0.1,right=1,top=0.9,wspace=0.2,hspace=0.2)
        plot_mod(fig)
        fig.savefig('%s%s.png'%(fig_root,name),dpi=128,format='png',bbox_inches='tight')
        pyplot.close(fig)
    def _parse_dwrt(self, dwrt):
        if dwrt is None: dwrt = {col:0 for col in self.xcols}
        if isinstance(dwrt,str):
            assert dwrt in self.xcols
            dwrt = {dwrt:1}
        if isinstance(dwrt,list) or isinstance(dwrt,ndarray):
            dwrt = atleast_1d(dwrt)
            assert dwrt.ndim==1 and len(dwrt)==self.d
            dwrt = {col:dwrt[i] for i,col in enumerate(self.xcols)}
        for xcol in self.xcols: dwrt[xcol] = dwrt[xcol] if xcol in dwrt else 0
        dwrt = atleast_1d([dwrt[col] for col in self.xcols]).astype(int)
        assert dwrt.ndim==1 and len(dwrt)==self.d
        return dwrt
    def _fit(self, x, y): 
        raise NotImplementedError
    def __call__(self, dfx, dwrt=None):
        assert dfx.ndim==2
        dwrt = self._parse_dwrt(dwrt)
        if isinstance(dfx,ndarray): dfx = pd.DataFrame(dfx,columns=self.xcols)
        dftf = pd.DataFrame({col:self.tf_funs[col](dfx[col].to_numpy()) for col in self.xcols})
        xtf = dftf[self.xcols].to_numpy()
        yhattf = self._eval_with_grad(xtf,dwrt=zeros(self.d))
        yhat = self.tf_funs[self.ycol+'.inv'](yhattf)
        if (dwrt==0).all(): return yhat
        # handle derivitives
        assert (dwrt>=0).all()
        if sum(dwrt)!=1: raise Exception("currently only supports 1 derivitive at a time")
        yhattfgrad = self._eval_with_grad(xtf,dwrt=dwrt)
        dwrt = self.xcols[argmax(dwrt)]
        # chain rule
        # t_j = T_j(x_j) for j=1,...,d
        # ytf = M(t_1,...,t_d)
        # y = T^{-1}(ytf)
        # dy/dx_i = M'(t_1,...,t_d)*T_j'(x_j)/T'(T^{-1}(ytf))
        dyhat_dxi = yhattfgrad*self.tf_funs[dwrt+'.prime'](dfx[dwrt].to_numpy())/self.tf_funs[self.ycol+'.prime'](yhattf)
        return dyhat_dxi
    def _eval_with_grad(self, x, dwrt):
        raise NotImplementedError
    def get_metrics(self,df):
        y = df[self.ycol].to_numpy()
        yhat = self.__call__(df)
        eps_abs = abs(yhat-y)
        metrics = {
            'RMSE Abs': sqrt(mean(eps_abs**2)),
            'MAE Abs': mean(eps_abs),
            'Max Abs': max(eps_abs)}
        return metrics
    def __repr__(self):
        s = '%s(%s)\n'%(self.name,','.join(self.xcols))
        s += '\ttrain data description\n\t\t%s'%str(self.descrip).replace('\n','\n\t\t')
        s += '\n\ttrain data metrics\n'
        for metric,val in self.train_metrics.items(): s += '\t\t%25s: %.2e\n'%(metric,val)
        return s

class CubicSpline1D_FI(FriendlyInterpolant):
    """ https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.CubicSpline.html#scipy.interpolate.CubicSpline """
    def _fit(self, x, y):
        assert x.shape[1]==1
        x = x.squeeze()
        ord = argsort(x)
        xsort = x[ord]
        ysort = y[ord]
        self.model = scipy.interpolate.CubicSpline(xsort,ysort,extrapolate=True)
    def _eval_with_grad(self, x, dwrt):
        y = self.model(x.squeeze(),dwrt[0])
        return y

class LinearAndNearestNeighbor_FI(FriendlyInterpolant):
    def _fit(self, x, y):
        assert x.shape[1]>1
        self.inhull_model = scipy.interpolate.LinearNDInterpolator(x,y,rescale=True)
        self.outhull_model = scipy.interpolate.NearestNDInterpolator(x,y,rescale=True)
    def _eval_with_grad(self, x, dwrt):
        assert (dwrt==0).all() # derivitives not supported by LinearAndNearestNeighbor_FI
        y = self.inhull_model(x)
        y[isnan(y)] = self.outhull_model(x[isnan(y)])
        return y

class SmootheSpline2D_FI(FriendlyInterpolant):
    def _fit(self, x, y):
        assert x.shape[1]==2 # SmootheSpline2D_FI only supports 2D input features
        self.model = scipy.interpolate.SmoothBivariateSpline(x=x[:,0],y=x[:,1],z=y)
    def _eval_with_grad(self, x, dwrt):
        y = self.model(x=x[:,0],y=x[:,1],dx=dwrt[0],dy=dwrt[1],grid=False)
        return y

if __name__ == '__main__':
    from matplotlib import pyplot
    import pandas as pd
    
    # 1D example
    x = linspace(0,1,16)[1:-1]**2
    y = sin(2*pi*x)/cos(pi*x/2)
    cs1d = CubicSpline1D_FI(
        df = pd.DataFrame({'x':x,'y':y}),
        ycol = 'y',
        tf_funs = {
            'x':lambda x:sqrt(x), 'x.prime':lambda x:1/(2*sqrt(x)),
            'y':lambda y:exp(y), 'y.prime':lambda y:exp(y), 'y.inv':lambda y:log(y)},
        name = 'cs1d',
        plot = [None,'x'])
    print(cs1d)
    yhat = cs1d(pd.DataFrame({'x':random.rand(3)**2}),dwrt={'x':1})
    
    # 2D examples
    x = 10**random.rand(8**2,2)
    y = sin(2*pi*x[:,0])*cos(2*pi*x[:,1])
    df = pd.DataFrame({'x_0':x[:,0],'x_1':x[:,1],'y':y})
    tf_funs = {
        'x_0':lambda x:log10(x), 'x_0.prime':lambda x:1/(x*log(10)),
        'x_1':lambda x:log10(x), 'x_1.prime':lambda x:1/(x*log(10))}
    lnn2d = LinearAndNearestNeighbor_FI(df,tf_funs,name='lnn2d',plot=None)
    print(lnn2d)
    yhat = lnn2d(pd.DataFrame(10**random.rand(3,2),columns=['x_0','x_1']))
    ss2d = SmootheSpline2D_FI(df,tf_funs,name='ss2d',plot=[None,'x_0','x_1'])
    yhat = ss2d(pd.DataFrame(10**random.rand(3,2),columns=['x_0','x_1']),dwrt='x_1')
    print(ss2d)

    # Arbitrary dimension example
    d = 5
    x = random.rand(8**3,d)
    y = prod(sin(2*pi*x),1)
    lnnd = LinearAndNearestNeighbor_FI(hstack([x,y[:,None]]),name='lnn%d'%d)
    print(lnnd)
    yhat = lnnd(random.rand(3,d))


    


import latexify
import numpy as np
from scipy import optimize
import scipy.integrate as integr

def zeros_func(mmin=5, mmax=1e9, beta=2.3):
    return lambda x: (1-beta)/(2-beta) * np.divide(x**(2-beta) - mmin**(2-beta), mmax**(1-beta) - x**(1-beta))

def minimize(func=zeros_func, total=2e7, *args, **kwargs):
    return lambda x: total - func(x, *args, **kwargs)



def mass_weighted_func(M, func, *args, **kwargs):
    '''Mass-weighted function'''
    return np.multiply(M, func(M, *args, **kwargs))

def plaw(M, neg_exponent=2.3):
    '''Decaying power law'''
    #if M > 0.:
    return M**(-neg_exponent)
    #else:
    #    print('Please provide positive masses')

def int_plaw(ll, ul, power):
    r'''
    Integration of the decaying power law
    
    $\int^{u}_{l}x^{\alpha} {\rm d}x = \frac{u^{1-\alpha} - l^{1-\alpha}}{1-\alpha}$
    '''
    if np.logical_or(power==1, power==2):
        return np.log(ul) - np.log(ll)
    else:
        return np.divide(plaw(ul, power-1) - plaw(ll, power-1), 1-power)

def broken_plaw_norm(limit, lower_exp, upper_exp):
    '''Ensure continuity in the broken power law'''
    return limit**(upper_exp - lower_exp)
    
def Kroupa01(Mstar, alpha0=0.3, alpha1=1.3, alpha2=2.3, alpha3=2.3,
            lim01=0.08, lim12=0.5, lim23=1., Ml=0.07, Mu=150):
    '''
    Broken power law. The default parameters are equivalent to 
    Kroupa (2001)
    https://ui.adsabs.harvard.edu/abs/2001MNRAS.322..231K/abstract 
    
    lim refer to the mass limits that break the power law
    '''
    a0 = 1.
    a1 = a0 * broken_plaw_norm(lim01, alpha0, alpha1)
    a2 = a1 * broken_plaw_norm(lim12, alpha1, alpha2)
    a3 = a2 * broken_plaw_norm(lim23, alpha2, alpha3)
    return np.piecewise(Mstar, 
            [
                np.logical_or(Mstar < Ml, Mstar >= Mu),
                np.logical_and(Mstar >= Ml, Mstar < lim01),
                np.logical_and(Mstar >= lim01, Mstar < lim12),
                np.logical_and(Mstar >= lim12, Mstar < lim23),
                np.logical_and(Mstar >= lim23, Mstar < Mu)
            ],
            [
                0., 
                lambda M: a0 * plaw(M, neg_exponent=alpha0),
                lambda M: a1 * plaw(M, neg_exponent=alpha1), 
                lambda M: a2 * plaw(M, neg_exponent=alpha2), 
                lambda M: a3 * plaw(M, neg_exponent=alpha3)
            ])

def integral_plaw(ll, ul, power):
    ''''''
    if ll < ul:
        return int_plaw(ll, ul, power)
    else:
        return 0.
        print('in integral_plaw, ll>=ul')

def normalized(x, func, condition:float=None, *args, **kwargs):
    '''Checks whether or not the mass function has been normalized'''
    if condition:
        if x <= condition:
            return func
        else:
            return 0.
    else:
        return func

def get_norm(normalization, IMF, Mtot:float, min_val:float, max_val:float, *args, **kwargs):
    '''get normalization, where normalization and IMF are both functions '''
    k, M = normalization(IMF, Mtot, min_val, min_val, *args, **kwargs)
    IMF_func = lambda mass: k * IMF(mass, m_max=M, *args, **kwargs)
    IMF_weighted_func = lambda mass: weighted_func(mass, IMF_func,
                                                   *args, **kwargs)
    return k, M, np.vectorize(IMF_func), np.vectorize(IMF_weighted_func)

def normalization_ECMF(ECMF_func, beta, Mtot, lower_lim, upper_lim, Mecl_max_guess=1e1,*args) -> (float, float):
    #print(f'{beta=}, \t{upper_lim=}')
    k_ECMF = lambda x: np.divide(1-beta, np.abs(upper_lim)**(1-beta) - np.abs(x)**(1-beta))
    #def k_ECMF(x):
    #    return np.reciprocal(integral_plaw(x, upper_lim, beta))
    #def k_ECMF(x):
    #    if np.logical_and(x >= lower_lim, x < upper_lim):
    #        print(f'{x=}')
    #        return np.reciprocal(integral_plaw(x, upper_lim, beta))
    #    else:
    #        return 0.
    mass_weighted_ECMF = lambda M_ecl: integral_plaw(lower_lim, upper_lim, beta-1)#mass_weighted_func(M_ecl, ECMF_func)
    func = lambda x: (k_ECMF(x) * mass_weighted_ECMF(x) - Mtot)
    #func = minimize(func=np.multiply(k_ECMF, mass_weighted_ECMF), total=Mtot)
    #print(f'{Mtot=:.2e}')
    #print(f'{lower_lim=:.2e}')
    #print(f'{upper_lim=:.2e}')
    #print(f'{k_ECMF(lower_lim)=:.2e}')
    #print(f'{k_ECMF(lower_lim) * mass_weighted_ECMF(lower_lim)=:.2e}')
    sol = optimize.root_scalar(func, x0=Mecl_max_guess, x1=1e8, rtol=1e-15)
    Mecl_max = sol.root
    #sol = optimize.root(func, Mecl_max_guess, method='hybr')
    #Mecl_max = sol.x[0]
    #print(f'{k_ECMF(Mecl_max)=:.2e}')
    #print(f'{sol.x=}')
    return k_ECMF(Mecl_max), Mecl_max

def normalization_IMF(alpha1, alpha2, alpha3, Mtot, lower_lim, upper_lim
                      ) -> (float, float):
    def k(x):
        if np.logical_and(x >= lower_lim, x < 0.5):
            integral = (2 * integral_plaw(x, 0.5, alpha1) 
                        + integral_plaw(0.5, 1., alpha2) 
                        + integral_plaw(1., upper_lim, alpha3))
            return np.divide(1, integral, where=integral!=0)
        if np.logical_and(x >= 0.5, x < 1.):
            integral = (integral_plaw(x, 1., alpha2) 
                                + integral_plaw(1., upper_lim, alpha3))
            return np.divide(1, integral, where=integral!=0)
        if np.logical_and(x >= 1., x <= upper_lim):
            integral = integral_plaw(x, upper_lim, alpha3)
            return np.divide(1, integral, where=integral!=0)
        else:
            return 0.
    def weighted_IMF(x):
        if np.logical_and(x >= lower_lim, x < 0.5):
            return (2 * integral_plaw(0.08, x, alpha1-1))
        if np.logical_and(x >= 0.5, x < 1.):
            return (2 * integral_plaw(0.08, 0.5, alpha1-1)
                    + integral_plaw(0.5, x, alpha2-1))
        if np.logical_and(x >= 1., x <= upper_lim):
            return (2 * integral_plaw(0.08, 0.5, alpha1-1)
                    + integral_plaw(0.5, 1., alpha2-1)
                    + integral_plaw(1., x, alpha3-1))
        else:
            return 0.
    func = lambda x: (k(x) * weighted_IMF(x) - Mtot)
    try:
        sol = optimize.root_scalar(func, method='bisect', rtol=1e-15,
                                   bracket=(lower_lim, upper_lim))
        m_max = sol.root
        #print(f'{sol.converged=}')
        #sol = optimize.root(func, 10, method='hybr')
        #m_max = sol.x[0]
    except:
        m_max = upper_lim
    #print(f'{Mtot = },\t {m_max = },\t {k(m_max)=}')
    return k(m_max), m_max

def find_closest_prod(number):
    '''
    Auxiliary function for subplots.
    Based on a Vector class resolution (number), adapts the number of subplots
    to fit all of the computed curves. Finds the number of rows and columns 
    (nl, nu) closest to a square that best accomodates the Vector resolution.
    
    number    Integer representing the resolution of the Vector class
    '''
    nl = int(np.floor(np.sqrt(number)))
    if number == np.sign(nl)*np.abs(nl)**2:
        nu = nl
    else:
        nu = nl+1
    while number > nu*nl:
        nu += 1
    return nl, nu


class Downsizing:
    """
    Downsizing relations for ellipticals 
    as it is introduced by Thomas et al. (2005)
    https://ui.adsabs.harvard.edu/abs/2005ApJ...621..673T/abstract
    """
    def __init__(self, M_igal: float) -> None:
        '''
        M_igal             [Msun] 
        downsizing_time    [yr]
        SFR                [Msun/yr]
        '''
        self.M_igal = M_igal
        self.downsizing_time = self.delta_tau(M_igal)
        self.SFR = self.SFR_func(self.M_igal, self.downsizing_time)
        
    def delta_tau(self, M_igal):
        '''
        Returns delta tau in Gyr for the downsizing relation 
        as it is expressed in Recchi+09
        
        M_igal is expressed in Msun and ranges from 1e6 to 1e12
        '''
        return 8.16 * np.e**(-0.556 * np.log10(M_igal) + 3.401) + 0.027       
            
    def SFR_func(self, M_igal, downsizing_time):
        '''SFR [Msun/yr] assuming the downsizing time (Thomas et al., 2005)'''
        return np.divide(M_igal, downsizing_time * 1e9)


import glob
import pickle
import dill
import itertools
import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
from igimf import plots as plts
#from create_interpolant import LinearAndNearestNeighbor_GCI
from igimf.friendly_interpolants import LinearAndNearestNeighbor_FI
plots = plts.Plots()

txts = glob.glob('igimf/grid/resolution15/*pkl')
#txts.remove('.DS_Store')
#txts.remove('resolution50')
df = pd.DataFrame({col:[] for col in ['SFR', 'metal_mass_fraction',
                                      'mass_star', 'IGIMF']})

print('Importing pickle files')
for txt in txts:
    df_txt = pickle.load(open(txt, 'rb'))
    df = pd.concat([df,df_txt])
print('Pickle files imported')
#df_grids = copy(df)

SFR_v = np.unique(df['SFR'])
metal_mass_fraction_v = np.unique(df['metal_mass_fraction'])
mstar_v = np.unique(df['mass_star'])

SFR_v_fit = np.logspace(np.log10(np.min(SFR_v)), np.log10(np.max(SFR_v)))
mstar_v_fit = np.logspace(np.log10(np.min(mstar_v)), np.log10(np.max(mstar_v)))
metal_mass_fraction_v_fit = np.logspace(np.log10(np.min(metal_mass_fraction_v)), np.log10(np.max(metal_mass_fraction_v)))

# print('Creating interpolation')
# IGIMF_interp = LinearAndNearestNeighbor_GCI(
#     df=df,
#     ycol='IGIMF',
#     tf_funs={
#         'metal_mass_fraction':lambda x:np.log10(x), 'metal_mass_fraction_prime':lambda x:1/(x*np.log(10)),
#         #'Mecl':lambda x:np.log10(x), 'Mecl_prime':lambda x:1/(x*np.log(10)),
#         'SFR':lambda x:np.log10(x), 'SFR_prime':lambda x:1/(x*np.log(10)),
#         'IGIMF':lambda y:np.log10(y), 'IGIMF_inv':lambda y:10**y, 'IGIMF_prime':lambda y:1/(y*np.log(10))
#     }, 
#     name='IGIMF')
# print(IGIMF_interp)
# print('Interpolation created')
#dill.dump(IGIMF_interp,open('IGIMFinterpolant.pkl','wb'))

#get_IGIMF = IGIMF_interp(pd.DataFrame({'metal_mass_fraction':0.0134, 'SFR':1}))
#IGIMF = get_IGIMF(mass_star)

'''
fi_igimf_full = LinearAndNearestNeighbor_FI(
    df = df,
    tf_funs = {},
    xcols = ['SFR','metal_mass_fraction','mass_star'],
    ycol = 'IGIMF',
    name = 'fi_igimf_full')
print(fi_igimf_full)
'''
fi_igimf_full = dill.load(open('igimf/fi_igimf_full.dill', 'rb'))

SFR_v = np.unique(df['SFR'])
metal_mass_fraction_v = np.unique(df['metal_mass_fraction'])
mstar_v = np.unique(df['mass_star'])


# Retrieve a value
fi_igimf_marginal = lambda mass_star: fi_igimf_full(pd.DataFrame({
    'mass_star': np.array(mass_star),
    'metal_mass_fraction': 0.0134,
    'SFR': 1}))
mass_star_query = df[(df['SFR']==df['SFR'][0])&(df['metal_mass_fraction']==df['metal_mass_fraction'][0])]['mass_star']
yhat = fi_igimf_marginal(mass_star=mass_star_query)
print(yhat)
import sys; sys.exit(0)


#Retrieve a value
df.loc[(df['SFR']==SFR_v[0]) & 
       (df['metal_mass_fraction']==metal_mass_fraction_v[0])]
       
#plots.sIMF_subplot(metal_mass_fraction_v, self.Mecl_v, self.mstar_v, self.IMF_Z_v_list)
## No IMF for SFR < 5e-5, i.e. Migal ~<6e5
fig1, ax1 = plt.subplots(1,1, figsize=(7,5))
for m in metal_mass_fraction_v:
    #num_colors=len(metal_mass_fraction_v)
    #currentColors = [cm(1.*i/num_colors) for i in range(num_colors)]
    #currentColor = itertools.cycle(currentColors)
    grid_sel = df.loc[(df['SFR']==SFR_v[10]) & (df['metal_mass_fraction']==m)]
    ax1.loglog(grid_sel['mass_star'], grid_sel['IGIMF'])
plt.show(block=False)
#fig1.savefig('figs/IGIMFbyZ_fixedSFR.pdf')

fig2, ax2 = plt.subplots(1,1, figsize=(7,5))  
for s in SFR_v:
    for m in metal_mass_fraction_v:
        grid_sel = df.loc[(df['SFR']==s) & (df['metal_mass_fraction']==m)]
        ax2.loglog(grid_sel['mass_star'], grid_sel['IGIMF'])
plt.show(block=False)
#fig2.savefig('figs/IGIMFbyZ_andbySFR.pdf')
       
fig3, ax3 = plt.subplots(1,1, figsize=(7,5))
for s in SFR_v:
    #num_colors=len(metal_mass_fraction_v)
    #currentColors = [cm(1.*i/num_colors) for i in range(num_colors)]
    #currentColor = itertools.cycle(currentColors)
    grid_sel = df.loc[(df['SFR']==s) & (df['metal_mass_fraction']==metal_mass_fraction_v[10])]
    ax3.loglog(grid_sel['mass_star'], grid_sel['IGIMF'])
plt.show(block=False)
#fig3.savefig('figs/IGIMFbySFR_fixedZ.pdf')


# print('Plotting IGIMF fit')
# fig4, ax4 = plt.subplots(1,1, figsize=(7,5))  
# for j,SFR_v_i in enumerate(SFR_v_fit):
#     for k, Z_v_i in enumerate(metal_mass_fraction_v_fit):
#         IGIMF_hats = np.zeros_like(mstar_v_fit)
#         for i,mstar_v_i in enumerate(mstar_v_fit):
#             df_test = pd.DataFrame({'SFR':[SFR_v_i], 'metal_mass_fraction':[Z_v_i],'Mecl':[10**5],'mass_star':[mstar_v_i]})
#             IGIMF_hats[i] = IGIMF_interp(df_test)
#         ax4.loglog(mstar_v_fit,IGIMF_hats,'-o')
# plt.show(block=False)
# print('Plotting of IGIMF fit completed')
#fig4.savefig('figs/IGIMFbyZ_fixedSFR.pdf')



print('Plotting of the 3D figure')
#plots.IGIMF_3D_plot(df, SFR_v, metal_mass_fraction_v, mstar_v, by_v='SFR', col_ax_idx=10)
plots.IGIMF_3D_plot(df, SFR_v, metal_mass_fraction_v, mstar_v, by_v='SFR', col_ax_idx=10, azim_rot=-120, elev_rot=20)
#plots.IGIMF_3Dlines_plot(df, SFR_v, metal_mass_fraction_v, mstar_v)
print('3D figure plotted')





import pathlib

import setuptools

setuptools.setup(
    name='igimf',
    version='0.0.1',
    description='Integrated Galaxy Wide IMF python package',
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type='text/markdown',
    url="https://github.com/egjergo/pyIGIMF",
    author='Eda Gjergo',
    author_email='eda.gjergo@gmail.com',
    license='LICENSE',
    project_urls={
        "Documentation": "https://readthedocs.io/pyIGIMF",
        "Source": "https://github.com/egjergo/pyIGIMF",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires=">3.8,<3.12",
    install_requires=["requests", "pandas>1.0"],
    # extras_require={
        
    # }
    packages=setuptools.find_packages(),
    include_package_data=True,
)


"""
Based on Gjergo, Kroupa, Yan, Sorokin, Jerabkova, and Zhang (2023, ApJS)

The Parameters superclass is parent to "ECMF", "StellarIMF", and "IGIMF".
    |... It contains the input parameters shared by all subclasses.

The ECMF subclass is structured as follows:
    |-- ECMF (normalized)
    |-- |-- embedded_cluster_MF (not normalized)
    |-- |-- |-- beta_func

The StellarIMF subclass is structured as follows:
    |-- stellar_IMF (normalized)
    |-- |-- initial_mass_function (not normalized)
    |-- |-- |-- alpha_1_func
    |-- |-- |-- alpha_2_func
    |-- |-- |-- alpha_3_func
    |-- |-- |-- |-- rho_cl
    |-- |-- |-- |-- x_alpha_3

The IGIMF subclass:
    |-- gwIMF_integrand_func, the integrand
    |-- gwIMF (output)
"""
import numpy as np
from scipy import optimize
import scipy.integrate as integr
import pandas as pd
import igimf.util as util

class Parameters:
    '''
    Parameters employed in all the subclasses
    
    INPUT
        metal_mass_fraction    [dimensionless] initial mass fraction 
                                of metals of the embedded cluster.
        SFR                    [Msun/yr] star formation rate in the galaxy.
        
    DEFAULT PARAMETERS
        solar_metallicity    [dimensionless] M_Z_sun/M_sun (Asplund+09)
        metallicity          [dimensionless] [Z]
        delta_alpha          [dimensionless] (page 3, Yan et al. 2021)
        m_star_max           [Msun] stellar mass upper limit, Yan et al. (2017)
        m_star_min           [Msun] stellar mass lower limit, Yan et al. (2017)
        M_ecl_max            [Msun] most-massive ultra-compact-dwarf galaxy
        M_ecl_min            [Msun] I've taken the lower limit from Eq. (8)!!!!
        delta_t              [yr] duration of the SF epoch
        '''
    def __init__(self, metal_mass_fraction: float, SFR: float,
                solar_metallicity=0.0134, delta_alpha=63.,
                m_star_max = 150.00001, m_star_min=0.07, suppress_warnings=False,
                M_ecl_max = 1e9, M_ecl_min=5., delta_t=1e7):
        vars = locals() 
        self.__dict__.update(vars)
        del self.__dict__["self"] 
        self.SFR = SFR 
        self.Mtot = self.SFR * self.delta_t 
        self.metal_mass_fraction = metal_mass_fraction
        self.metallicity = np.log10(self.metal_mass_fraction
                                    / self.solar_metallicity)
        if suppress_warnings:
            import warnings
            warnings.filterwarnings('ignore')


class ECMF(Parameters):
    ''' Embedded Cluster Mass Function
    Only depends on SFR, but needs metal_mass_fraction to import Parameters,
    (this preserves a consistency when called inside IGIMF,
    but if an ECMF is needed by itself, pass it a dummy metal_mass_fraction)
    
    i.e., evaluate ECMF at time t where the galaxy is characterized by
    a SFR(t) and a Z(t) -- but only SFR(t) affects the ECMF.
    '''
    def __init__(self, SFR:float=None, metal_mass_fraction:float=None):
        super().__init__(metal_mass_fraction, SFR)
        self.beta_ECMF = self.beta_func()
        self.call_ECMF()

    def beta_func(self):
        """Eq. (11) ECMF slope"""
        return -0.106 * np.log10(self.SFR) + 2.
    
    def ECMF_func(self, M_ecl, beta):
        util.plaw(M_ecl, neg_exponent=beta)
        return 
    
    def embedded_cluster_MF(self, M_ecl, m_max=None):
        r"""Eq. (8) ECMF (not normalized)"""
        if M_ecl>=self.M_ecl_min:
           return util.normalized(M_ecl, M_ecl**(-self.beta_ECMF), 
                                  condition=m_max)
        else:
            return 0.
               
    def call_ECMF(self):
        '''ECMF (normalized)'''
        self.k_ecl, self.M_max = util.normalization_ECMF(
                        self.embedded_cluster_MF,
                        self.beta_ECMF, self.SFR * self.delta_t, 
                        self.M_ecl_min, self.M_ecl_max)
        self.k_ecl = k_ecl
        self.M_max = M_max
        ECMF_func = lambda M_ecl: (k_ecl *
                                self.embedded_cluster_MF(M_ecl, m_max=M_max))
        ECMF_weighted_func = lambda M_ecl: util.weighted_func(M_ecl, ECMF_func)
        self.ECMF_func = np.vectorize(ECMF_func)
        self.ECMF_weighted_func = np.vectorize(ECMF_weighted_func)
    

class StellarIMF(Parameters):
    ''' Computes Initial Mass Function for an Embedded cluster (e.cl.)
    at a given time t where the e.cl. is characterized by a SFR(t) and a Z(t)
    
    Depends on the total mass and metallicity of the e.cl., and implicitly
    (through the M_ecl) on the SFR.
    
    RETURNS
        all parameters that characterize the stellar IMF,
        the stellar IMF function,
        the mass-weighted stellar IMF function.
    '''
    def __init__(self, M_ecl:float, metal_mass_fraction: float, SFR: float):
        super().__init__(metal_mass_fraction, SFR)
        self.M_ecl = M_ecl
        self.alpha_1 = float(self.alpha_1_func())
        self.alpha_2 = float(self.alpha_2_func())
        self.alpha_3 = self.alpha_3_func()
        self.stellar_IMF()
        self.IMF_func = np.vectorize(self.IMF_func)
        self.IMF_weighted_func = np.vectorize(self.IMF_weighted_func)

    def alpha_1_func(self):
        r"""Eq. (4) pt.1"""
        return (1.3 + self.delta_alpha * (self.metal_mass_fraction
                                          - self.solar_metallicity))
    
    def alpha_2_func(self):
        r"""Eq. (4) pt. 2 $\alpha_2 - \alpha_1 = 1$ always holds"""
        return 1 + self.alpha_1
        
    def rho_cl(self):
        r"""Eq. (7) core density of the molecular cloud 
        which forms the embedded star cluster
        In units of [Mstar/pc$^3$]
    
        For example, for M_ecl = 1000 Msun:
        >>> rho_cl(10**3)
        gives a core density of 4.79e4 Msun/pc$^3$"""
        return 10**(0.61 * np.log10(self.M_ecl) + 2.85)
    
    def x_alpha_3_func(self):
        r"""Eq. (6)"""
        return (-0.14 * self.metallicity + 0.99 * np.log10(self.rho_cl()/1e6))

    def alpha_3_func(self):
        r"""Eq. (5)"""
        x_alpha_3 = self.x_alpha_3_func()
        if x_alpha_3 < -0.87:
            return 2.3
        else:
            return -0.41 * x_alpha_3 + 1.94
        
    def initial_mass_function(self, m, alpha_3=None, m_max=None):
        util.Koupa01(m, alpha1=self.alpha1, alpha2=self.alpha2, alpha3=alpha3,
            lim12=0.5, lim23=1., Ml=0.07, Mu=150.)
        '''stellar IMF (not normalized)'''
        if np.logical_and(m>=self.m_star_min, m<0.5):
            return m**(-self.alpha_1) * 2
        elif np.logical_and(m>=0.5, m<1.):
            return m**(-self.alpha_2)
        elif m>=1.:
            return util.normalized(m, m**(-alpha_3), condition=m_max)
        else:
            return 0.
        
    def stellar_IMF(self):
        r"""Eq. (1) Returns the normalized stellar IMF xi: 
        $\xi_* = d N_* / d m $"""
        self.k_star, self.m_max = util.normalization_IMF(self.alpha_1, 
                        self.alpha_2, self.alpha_3, self.M_ecl, 
                        self.m_star_min, self.m_star_max)
        self.IMF_func = lambda m: self.k_star * self.initial_mass_function(m, 
                                        alpha_3=self.alpha_3, m_max=self.m_max)
        self.IMF_weighted_func = lambda m: util.weighted_func(m, self.IMF_func)

        
class IGIMF(Parameters):
    ''' Computes the Integrated Galaxy-wide Initial Mass Function of stars
    at a given time t, where the galaxy is characterized by 
    
    a metal mass fraction Z(t) = M_Z/M_gas, and
    a star formation rate SFR(t) [Msun/yr]
    
    RETURNS
        IGIMF    (function(mstar)) [#/Msun] the galaxy-wide IMF function
                 from the IGIMF formalism. The IGIMF function accepts
                 one variable, the mstar (either float or array)
    '''
    def __init__(self, metal_mass_fraction: float, SFR: float, 
                 suppress_warnings=True) -> None:
        super().__init__(metal_mass_fraction, SFR)
        self.IGIMF_func = np.vectorize(self.gwIMF())

    def gwIMF_integrand_func(self, M_ecl, ECMF_func, m):
        stellar_IMF = StellarIMF(M_ecl, self.metal_mass_fraction, self.SFR)
        return stellar_IMF.IMF_func(m) * ECMF_func(M_ecl)
    
    def gwIMF(self):
        r"""Eq. (12)"""
        ecl_MF = ECMF(self.SFR, metal_mass_fraction=self.metal_mass_fraction)
        return lambda m: integr.quadrature(self.gwIMF_integrand_func, 
            self.M_ecl_min, ecl_MF.M_max, args=(ecl_MF.ECMF_func, m), 
            vec_func=False, rtol=1e-15)[0]  
    
def main():
    import json
    massfrac = float(input("What is the metal mass fraction?") or .1*0.0134)
    SFR = float(input(r"What is the SFR (in $M_{\odot}$/yr)?") or 1)
    Mtot = float(input('''What is the mass of the embedded cluster?
    Note: this computes a single instance of the stellar IMF''') or 1e10)
    stellar_IMF = StellarIMF(Mtot, massfrac, SFR)
    ecl_MF = ECMF(SFR,metal_mass_fraction=massfrac)
    print(f'\nAt a time t characterized by:\n{SFR=:.2e}\nand \n{Mtot=:.2e}\n')
    print(f"\ndictionary of an e.cl.'s stellar IMF\n")
    print(json.dumps(stellar_IMF.__dict__, default=str, indent=4))
    print(f"\n\ndictionary of a galaxy's ECMF\n")
    print(json.dumps(ecl_MF.__dict__, default=str, indent=4))
    
    t_IGIMF = IGIMF(massfrac, SFR)
    print(f"\n\ndictionary of a galaxy's IGIMF at a given time t\n")
    print(json.dumps(t_IGIMF.__dict__, default=str, indent=4))
    return massfrac, SFR, Mtot, stellar_IMF, ecl_MF, t_IGIMF
        
if __name__ == '__main__':
    massfrac, SFR, Mtot, stellar_IMF, ecl_MF, t_IGIMF = main()


import dill
import pandas as pd
import numpy as np

def create_igimf_fit():
    import glob
    import pickle
    import dill
    import itertools
    import pandas as pd
    import numpy as np
    from matplotlib import pyplot as plt
    from igimf import plots as plts
    from igimf.friendly_interpolants import LinearAndNearestNeighbor_FI
    plots = plts.Plots()

    txts = glob.glob('igimf/grid/resolution15/*pkl')
    #txts.remove('.DS_Store')
    #txts.remove('resolution50')
    df = pd.DataFrame({col:[] for col in ['SFR', 'metal_mass_fraction',
                                        'mass_star', 'IGIMF']})

    print('Importing pickle files')
    for txt in txts:
        df_txt = pickle.load(open(txt, 'rb'))
        df = pd.concat([df,df_txt])
    print('Pickle files imported')
    #df_grids = copy(df)

    SFR_v = np.unique(df['SFR'])
    metal_mass_fraction_v = np.unique(df['metal_mass_fraction'])
    mstar_v = np.unique(df['mass_star'])

    SFR_v_fit = np.logspace(np.log10(np.min(SFR_v)), np.log10(np.max(SFR_v)))
    mstar_v_fit = np.logspace(np.log10(np.min(mstar_v)), np.log10(np.max(mstar_v)))
    metal_mass_fraction_v_fit = np.logspace(np.log10(np.min(metal_mass_fraction_v)), np.log10(np.max(metal_mass_fraction_v)))

    fi_igimf_full = LinearAndNearestNeighbor_FI(
        df = df,
        tf_funs = {},
        xcols = ['SFR','metal_mass_fraction','mass_star'],
        ycol = 'IGIMF',
        name = 'fi_igimf_full')
    
    dill.dump(fi_igimf_full, open('igimf/fi_igimf_full_.dill', 'wb'))
    return None

def import_igimf_fit():
    return dill.load(open('igimf/fi_igimf_full.dill', 'rb'))

def input_igimf(SFR: float, metal_mass_fraction: float):
    '''
    INPUT
        SFR                    [Msun/yr] star formation in units of solar masses per year
        metal_mass_fraction    [ ] (M_Z/M_gas) ratio of mass in metals divided by gas mass
        
    OUTPUT
        implicit function of stellar mass
        contains a pandas dataframe, to be used as input  for the IGIMF interpolation
    
    EXAMPLE
    >>> import numpy as np
    >>> stellar_masses = np.logspace(0.07, 150)
    >>> IGIMF_import = dill.load(open('fi_igimf_full.dill', 'rb'))
    >>> IGIMF_in = IGIMF_input(SFR=2, metal_mass_fraction=0.01)(stellar_masses)
    >>> IGIMF_fit = IGIMF_import(IGIMF_in)
    '''
    return lambda mass_star: pd.DataFrame({
        'mass_star': np.array(mass_star),
        'metal_mass_fraction': metal_mass_fraction,
        'SFR': SFR
    })
    
def compute_IGIMF(SFR:float, metal_mass_fraction:float,
                  mstar_upperlim=150., mstar_lowerlim=0.07, mstar_vector_length=50):
    mstar_v_fit = np.logspace(np.log10(mstar_lowerlim), np.log10(mstar_upperlim), num=mstar_vector_length)
    IGIMF_import =import_igimf_fit()
    IGIMF_input = input_igimf(SFR, metal_mass_fraction)
    return IGIMF_import(IGIMF_input(mstar_v_fit))